package com.cloudwords.api.client;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.util.Collection;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import org.apache.commons.compress.archivers.zip.ZipArchiveEntry;
import org.apache.commons.compress.archivers.zip.ZipArchiveInputStream;
import org.apache.commons.compress.utils.IOUtils;
import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.HttpStatus;
import org.apache.http.client.ClientProtocolException;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpDelete;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.methods.HttpPut;
import org.apache.http.client.methods.HttpUriRequest;
import org.apache.http.entity.StringEntity;
import org.apache.http.entity.mime.MultipartEntityBuilder;
import org.apache.http.entity.mime.content.FileBody;

import com.cloudwords.api.client.exception.CloudwordsClientException;
import com.cloudwords.api.client.resources.CloudwordsFile;
import com.cloudwords.api.client.resources.CodeAndDisplay;
import com.cloudwords.api.client.resources.Error;
import com.cloudwords.api.client.resources.IntendedUse;
import com.cloudwords.api.client.resources.Language;
import com.cloudwords.api.client.resources.Vendor;
import com.cloudwords.api.client.serializers.CodeAndDisplaySerializer;
import com.cloudwords.api.client.serializers.DateSerializer;
import com.cloudwords.api.client.serializers.IntendedUseSerializer;
import com.cloudwords.api.client.serializers.LanguageSerializer;
import com.cloudwords.api.client.serializers.VendorSerializer;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.JavaType;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.module.SimpleModule;
import com.fasterxml.jackson.databind.type.TypeFactory;

/**
 * Base implementation of a Cloudwords API client.
 * 
 * @author jonathansu
 * @since 1.3
 */
public abstract class BaseCloudwordsClient {

    public static final String UTF_8 = "UTF-8";
    protected static final String JSON_CONTENT_TYPE = "application/json";
    protected static final String ZIP_CONTENT_TYPE = "application/zip";
    protected static final String CW_FILE_PARAM = "file";
    protected static final String CW_SOURCE_PARAM = "source";
    protected static final int connectionTimeoutMillis = 30000;
    protected static final int socketTimeoutMillis = 60000;
    protected static final int maxTotalConnections = 3;

    protected String baseUrlWithVersion;
    protected double apiVersion;
    
    protected abstract HttpClient getClient();
    protected abstract ObjectMapper getMapper();
    protected abstract String getAuthToken();
    
    public InputStream downloadFileFromMetadata(CloudwordsFile metadata) throws IllegalStateException,
            CloudwordsClientException, IOException {
        if (metadata != null && metadata.getContentPath() != null) {
            return streamFileFromPath(metadata.getContentPath());
        } else {
            return null;
        }
    }

    @SuppressWarnings("unchecked")
    protected <T extends Object> T executeRequest(HttpUriRequest request, Class<T> expectedResponseType)
            throws CloudwordsClientException {
        return (T) executeRequest(request, TypeFactory.defaultInstance().constructType(expectedResponseType));
    }

    protected <T extends Object> T executeRequest(HttpUriRequest request, TypeReference<T> expectedResponseType)
            throws CloudwordsClientException {
        return executeRequest(request, TypeFactory.defaultInstance().constructType(expectedResponseType));
    }
    
    protected <T extends Object> T executeRequest(HttpUriRequest request, JavaType expectedResponseType)
            throws CloudwordsClientException {
        HttpResponse response;
        BufferedReader in = null;
        try {
            response = getClient().execute(request);
            in = new BufferedReader(new InputStreamReader(response.getEntity().getContent()));

            int statusCode = response.getStatusLine().getStatusCode();
            if (isRequestSuccessful(statusCode)) {
                return getMapper().readValue(in, expectedResponseType);
            } else {
                Error err;
                try {
                    err = getMapper().readValue(in, Error.class);
                } catch (Exception ex) {
                    throw new CloudwordsClientException(ex);
                }
                err.setHttpStatusCode(statusCode);
                throw new CloudwordsClientException(err);
            }
        } catch (ClientProtocolException ex) {
            throw new CloudwordsClientException(ex);
        } catch (IOException ex) {
            throw new CloudwordsClientException(ex);
        } finally {
            if (in != null) {
                try {
                    in.close();
                } catch (IOException ex) {
                    throw new CloudwordsClientException(ex);
                }
            }
        }
    }
    
    protected <T> T get(String path, Class<T> type) throws CloudwordsClientException   {
        HttpGet httpGet = new HttpGet(baseUrlWithVersion + path);
        setBasicHeaders(httpGet, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        return executeRequest(httpGet, type);
    }
    
    protected <T> T get(String path, TypeReference<T> type, Object params) throws CloudwordsClientException  {
        if (params != null) {
            String query = buildQuery(params);
            if (query.length() != 0)    {
                path += "?" + query;
            }
        }
        
        HttpGet httpGet = new HttpGet(baseUrlWithVersion + path);
        setBasicHeaders(httpGet, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        return executeRequest(httpGet, type);
    }
    
    protected String buildQuery(Object params)  {
        try {
            Map<String,Object> map = getMapper().convertValue(params, new TypeReference<Map<String,Object>>() {});
            
            StringBuilder buf = new StringBuilder();
            
            for (Entry<String,Object> entry : map.entrySet())  {
                String key = entry.getKey();
                Object value = entry.getValue();
                
                if (value instanceof Collection)    {
                    for (Object item : ((Collection<?>) value))    {
                        if (item != null)   {
                            if (buf.length() > 0)   {
                                buf.append('&');
                            }
                            buf.append(URLEncoder.encode(key, UTF_8));
                            buf.append('=');
                            buf.append(URLEncoder.encode(item.toString(), UTF_8));
                        }
                    }
                } else if (value != null)   {
                    if (buf.length() > 0)   {
                        buf.append('&');
                    }
                    buf.append(URLEncoder.encode(key, UTF_8));
                    buf.append('=');
                    buf.append(URLEncoder.encode(value.toString(), UTF_8));
                }
            }
            
            return buf.toString();
        } catch (UnsupportedEncodingException e) {
            throw new RuntimeException(e);
        }
    }
    
    protected <T> List<T> getList(String path, Class<T> elemType) throws CloudwordsClientException {
        HttpGet httpGet = new HttpGet(baseUrlWithVersion + path);
        setBasicHeaders(httpGet, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        return executeRequest(httpGet, TypeFactory.defaultInstance().constructCollectionType(List.class, elemType));
    }
    
    @SuppressWarnings("unchecked")
    protected <T> T put(String path, T object) throws CloudwordsClientException  {
        HttpPut request = new HttpPut(baseUrlWithVersion + path);
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        try {
            request.setEntity(new StringEntity(getMapper().writeValueAsString(object), UTF_8));
        } catch (Exception ex) {
            throw new CloudwordsClientException(ex);
        }
        return (T) executeRequest(request, object.getClass());
    }
    
    protected <T> List<T> putList(String path, Class<T> elemType, List<T> objects) throws CloudwordsClientException  {
        HttpPut request = new HttpPut(baseUrlWithVersion + path);
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        try {
            request.setEntity(new StringEntity(getMapper().writeValueAsString(objects), UTF_8));
        } catch (Exception ex) {
            throw new CloudwordsClientException(ex);
        }
        return executeRequest(request, TypeFactory.defaultInstance().constructCollectionType(List.class, elemType));
    }
    
    protected <T> T postFile(String path, File file, Class<T> returnType) throws CloudwordsClientException {
        HttpPost request = new HttpPost(baseUrlWithVersion + path);
        
        HttpEntity entity = MultipartEntityBuilder.create().addPart(CW_FILE_PARAM, new FileBody(file)).build();
        request.setEntity(entity);

        setBasicHeaders(request, JSON_CONTENT_TYPE, entity.getContentType().getValue());
        return executeRequest(request, returnType);
    }
    
    protected HttpResponse postFile(String path, File file) throws CloudwordsClientException {
        HttpPost request = new HttpPost(baseUrlWithVersion + path);
        
        HttpEntity entity = MultipartEntityBuilder.create().addPart(CW_FILE_PARAM, new FileBody(file)).build();
        request.setEntity(entity);

        setBasicHeaders(request, JSON_CONTENT_TYPE, entity.getContentType().getValue());
        
        try {
        	return getClient().execute(request);
        } catch(Exception e) {
        	throw new CloudwordsClientException(e);
        }
    }
    
    protected <T> T putFile(String path, File file, Class<T> returnType) throws CloudwordsClientException {
        HttpPut request = new HttpPut(baseUrlWithVersion + path);

        HttpEntity entity = MultipartEntityBuilder.create().addPart(CW_FILE_PARAM, new FileBody(file)).build();
        request.setEntity(entity);

        setBasicHeaders(request, JSON_CONTENT_TYPE, entity.getContentType().getValue());
        return executeRequest(request, returnType);
    }
    
    protected void putFile(String path, File file) throws CloudwordsClientException {
        HttpPut request = new HttpPut(baseUrlWithVersion + path);

        HttpEntity entity = MultipartEntityBuilder.create().addPart(CW_FILE_PARAM, new FileBody(file)).build();
        request.setEntity(entity);

        setBasicHeaders(request, JSON_CONTENT_TYPE, entity.getContentType().getValue());
        executeRequest(request);
    }
    
    protected void delete(String path) throws CloudwordsClientException  {
        HttpDelete request = new HttpDelete(baseUrlWithVersion + path);
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        executeRequest(request);
    }
    
    protected HttpResponse executeRequest(HttpUriRequest request)
            throws CloudwordsClientException {
        HttpResponse response;
        BufferedReader in = null;
        try {
            response = getClient().execute(request);
            in = new BufferedReader(new InputStreamReader(response.getEntity().getContent()));
            int statusCode = response.getStatusLine().getStatusCode();
            if (!isRequestSuccessful(statusCode)) {
                Error err;
                try {
                    err = getMapper().readValue(in, Error.class);
                } catch (Exception ex) {
                    throw new CloudwordsClientException(ex);
                }
                err.setHttpStatusCode(statusCode);
                throw new CloudwordsClientException(err);
            }
            return response;
        } catch (ClientProtocolException ex) {
            throw new CloudwordsClientException(ex);
        } catch (IOException ex) {
            throw new CloudwordsClientException(ex);
        } finally {
            if (in != null) {
                try {
                    in.close();
                } catch (IOException ex) {
                    throw new CloudwordsClientException(ex);
                }
            }
        }
    }

    protected void setBasicHeaders(HttpUriRequest request, String acceptContentType, String bodyContentType) {
        request.setHeader("Accept", acceptContentType);
        request.setHeader("Content-type", bodyContentType);
        request.setHeader("Authorization", getAuthToken());
    }

    protected void initObjectMapper() {
        SimpleModule module = new SimpleModule();
        module.addSerializer(IntendedUse.class, new IntendedUseSerializer());
        module.addSerializer(Language.class, new LanguageSerializer());
        module.addSerializer(Date.class, new DateSerializer());
        module.addSerializer(Vendor.class, new VendorSerializer());
        module.addSerializer(CodeAndDisplay.class, new CodeAndDisplaySerializer());
        getMapper().registerModule(module);
        getMapper().configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);
    }

    protected boolean isRequestSuccessful(int statusCode) {
        return isErrorCodeMatched(statusCode, HttpStatus.SC_OK) || isErrorCodeMatched(statusCode, HttpStatus.SC_CREATED);
    }

    protected boolean isErrorCodeMatched(int statusCode, int codeToMatch) {
        if (statusCode == codeToMatch) {
            return true;
        }
        return false;
    }
    
    protected void validateZipFile(File zipFile) throws CloudwordsClientException {
        boolean isZip = false;
        ZipArchiveInputStream zis = null;
        try {
        	InputStream is = new FileInputStream(zipFile);
            zis = new ZipArchiveInputStream(is);
            ZipArchiveEntry zipEntry = zis.getNextZipEntry();
            isZip = (zipEntry != null);
        } catch (IOException ioEx) {
            isZip = false;
        } finally {
            if ( zis != null ) {
                try {
                    zis.close();
                } catch (IOException ioEx) {
                    // Passing on this error as this is not expected to be a common occurrence.
                    throw new IllegalStateException(ioEx);
                }
            }
        }
        
        if (!isZip) {
            throw new CloudwordsClientException("The provided file must be a valid zip archive");
        }
        
    }

    /**
     * Stream the requested file back to the http response. Please note that the input stream MUST
     * BE properly closed after use.
     * 
     * @throws IOException
     * @throws IllegalStateException
     */
    protected InputStream streamFileFromPath(String path) throws CloudwordsClientException, IllegalStateException,
            IOException {
        HttpUriRequest request = new HttpGet(path);
        setBasicHeaders(request, ZIP_CONTENT_TYPE, JSON_CONTENT_TYPE);

        HttpResponse response = getClient().execute(request);
        int statusCode = response.getStatusLine().getStatusCode();
        if (isRequestSuccessful(statusCode)) {
            return response.getEntity().getContent();
        } else {
            Error err;
            BufferedReader in = null;
            try {
                in = new BufferedReader(new InputStreamReader(response.getEntity().getContent()));
                err = getMapper().readValue(in, Error.class);
            } finally {
                if (in != null) {
                    try {
                        in.close();
                    } catch (IOException ex) {
                        throw new CloudwordsClientException(ex);
                    }
                }
            }
            err.setHttpStatusCode(statusCode);
            throw new CloudwordsClientException(err);
        }
    }
    
	/**
	 * Returns filename without extension.
	 * @param nameWithExtension
	 * @return
	 * 	Name without extension.
	 */
	protected static String getFilename(String nameWithExtension) {
		int dotIndex = nameWithExtension.lastIndexOf(".");
		if(dotIndex != -1) {
			return nameWithExtension.substring(0, dotIndex);
		} else {
			//no extension, so return complete name
			return nameWithExtension;
		}
	}

	/**
	 * Adds zip entry to passed output stream.
	 * @param zos
	 *   Zip output stream
	 * @param dirName
	 *   Destination zip directory
	 * @param fileToAdd
	 *   File in zip entry
	 * @throws IOException
	 */
	protected static void addZipEntry(ZipOutputStream zos, String dirName, File fileToAdd) throws IOException {
		FileInputStream in = null;
		try {
			in = new FileInputStream(fileToAdd);
			ZipEntry entry = new ZipEntry(dirName + "/" + fileToAdd.getName());
			zos.putNextEntry(entry);
			IOUtils.copy(in, zos);
			zos.closeEntry();
		} finally {
			if(in != null) {
				IOUtils.closeQuietly(in);
			}
		}
	}
    
}
