/*
 *
 * Licensed under the API LICENSE AGREEMENT, Version 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.cloudwords.com/developers/license-1.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.cloudwords.api.client;

import java.io.File;
import java.util.List;

import com.cloudwords.api.client.exception.CloudwordsClientException;
import com.cloudwords.api.client.resources.BidRequest;
import com.cloudwords.api.client.resources.ChangeOrderProjectWorkflow;
import com.cloudwords.api.client.resources.CloudwordsFile;
import com.cloudwords.api.client.resources.CodeAndDisplay;
import com.cloudwords.api.client.resources.CustomerBid;
import com.cloudwords.api.client.resources.CustomerChangeOrder;
import com.cloudwords.api.client.resources.CustomerChangeOrderBid;
import com.cloudwords.api.client.resources.CustomerTask;
import com.cloudwords.api.client.resources.CustomerTranslationRequest;
import com.cloudwords.api.client.resources.Department;
import com.cloudwords.api.client.resources.IntendedUse;
import com.cloudwords.api.client.resources.Language;
import com.cloudwords.api.client.resources.PagedProjectQuery;
import com.cloudwords.api.client.resources.PagedResponse;
import com.cloudwords.api.client.resources.Project;
import com.cloudwords.api.client.resources.ProjectWorkflow;
import com.cloudwords.api.client.resources.ReviewerInstructions;
import com.cloudwords.api.client.resources.SourceDocument;
import com.cloudwords.api.client.resources.Task;
import com.cloudwords.api.client.resources.TranslatedDocument;
import com.cloudwords.api.client.resources.TranslationRequest;
import com.cloudwords.api.client.resources.User;
import com.cloudwords.api.client.resources.Vendor;
import com.cloudwords.api.client.resources.WorkflowTemplate;

/**
 * Client interface for the Cloudwords API. For more details and best practices please refer to the
 * Cloudwords API documentation found at <a
 * href="http://www.cloudwords.com/developers">Cloudwords</a>.
 * <p>
 * This API is designed for Cloudwords customers that are looking to:
 * <ul>
 * <li>Create and update a Cloudwords {@link Project}</li>
 * <li>Upload and download of a {@link Project}'s source content and reference materials</li>
 * <li>Request bids for a {@link Project}</li>
 * <li>Download translated content (both by individual language and via the aggragated
 * "master deliverable" that contains all deliverables across all languages)</li>
 * </ul>
 * 
 * @author jonathansu
 * @since 1
 */
public interface CloudwordsCustomerAPI extends BaseCloudwordsAPI {

    

    /**
     * Create a new Cloudwords {@link Project}. Please refer to the API documentation for more
     * information.
     * 
     * @param project
     *            {@link Project} to be created
     * @return created {@link Project} resource
     * @throws CloudwordsClientException
     */
    public Project createProject(Project project) throws CloudwordsClientException;
    
    public List<Project> getOpenProjects(Department department) throws CloudwordsClientException;
    public List<Project> getClosedProjects(Department department) throws CloudwordsClientException;

    /**
     * Upload the source material for a given Cloudwords {@link Project}. The provided file must be
     * a zip file.
     * <p>
     * Please note that this {@link Project} must already exists in Cloudwords, and the
     * {@link Project} must be in a state where source material can be uploaded. If source material
     * for this {@link Project} already exists, then this operation will overwrite the source
     * material associated with the {@link Project}. Please refer to the API documentation for more
     * information.
     * </p>
     * 
     * @param projectId
     *            ID of {@link Project} to upload source for
     * @param zipFile
     *            A single zip file to be uploaded for this {@link Project}
     * @return Uploaded source material {@link CloudwordsFile} resource
     * @throws CloudwordsClientException
     */
    public CloudwordsFile uploadProjectSource(int projectId, File zipFile) throws CloudwordsClientException;
    
    /**
     * Add an individual source document. As of API 1.14, this is the preferred way of uploading project source. 
     * @param projectId The ID of the project
     * @param file The file to be translated
     * @return The metadata of the new source document
     * @throws CloudwordsClientException
     */
    public SourceDocument addSourceDocument(int projectId, File file) throws CloudwordsClientException;
    
    
    
    /**
     * Replace the content of a source document that is already attached to the project. This is only valid before a bid request has been created for a project.
     * @param projectId The project ID
     * @param documentId The document ID
     * @param file The new content
     * @return The updated metadata for the document
     * @throws CloudwordsClientException
     */
    public SourceDocument replaceSourceDocument(int projectId, int documentId, File file) throws CloudwordsClientException;
    
    /**
     * Delete a source document on a project. This is only valid before a bid request has been created for a project.
     * @param projectId 
     * @param documentId
     * @throws CloudwordsClientException
     */
    public void deleteSourceDocument(int projectId, int documentId) throws CloudwordsClientException;
    
    /**
     * Upload and add a new reference material for a given Cloudwords {@link Project}. The provided
     * file must be a zip file.
     * <p>
     * Please note that this {@link Project} must already exists in Cloudwords, and the
     * {@link Project} must be in a state where reference material can be uploaded. Please refer to
     * the API documentation for more information.
     * </p>
     * 
     * @param projectId
     *            ID of the {@link Project} to upload reference materials for
     * @param zipFile
     *            A single zip file to be uploaded for this {@link Project}
     * @return Uploaded reference material {@link CloudwordsFile} resource
     * @throws CloudwordsClientException
     */
    public CloudwordsFile uploadProjectReference(int projectId, File zipFile) throws CloudwordsClientException;

    /**
     * Upload and replace {@link Project} reference material for the specified Cloudwords
     * {@link Project} and specific reference material. The provided file must be a zip file.
     * <p>
     * Please note that this {@link Project} must already exists in Cloudwords, and the
     * {@link Project} must be in a state where reference material can be uploaded. Please refer to
     * the API documentation for more information.
     * </p>
     * 
     * @param projectId
     *            ID of the {@link Project} whose reference materials are to be updated
     * @param documentId
     *            ID of the specific reference material to update
     * @param zipFile
     *            A single zip file to be uploaded for this {@link Project}
     * @return updated reference material {@link CloudwordsFile} resource
     * @throws CloudwordsClientException
     */
    public CloudwordsFile updateProjectReference(int projectId, int documentId, File zipFile)
            throws CloudwordsClientException;

    
    /**
     * Request bids for the specified {@link Project}. In order to request a bid for a
     * {@link Project}, a source material must be already uploaded for the {@link Project}, and the
     * {@link Project} must be in a state where requesting bids is allowed.
     * <p>
     * Additionally, you must either specify the list of Cloudwords preferred {@link Vendor}(s) to
     * request bids from or to allow Cloudwords to automatically find {@link Vendor}s that best
     * match the {@link Project} requirements, or both.
     * </p>
     * Please refer to the API documentation for more information.
     * 
     * @param projectId
     *            ID of the {@link Project} to request bids for
     * @param preferredVendors
     *            The list of {@link Vendor}s to request bids from
     * @param doLetCloudwordsChoose
     *            Allow Cloudwords to find {@link Vendor} that best match the {@link Project}
     * @param doAutoSelectBidFromVendor
     *            Whether to automatically select the bid provided by the {@link Vendor} for this
     *            {@link BidRequest}
     * @return created {@link BidRequest} resource
     * @throws CloudwordsClientException
     */
    public BidRequest requestBidsForProject(int projectId, List<Vendor> preferredVendors, boolean doLetCloudwordsChoose, boolean doAutoSelectBidFromVendor)
            throws CloudwordsClientException;
    
    
    /**
     * Get the current {@link BidRequest} for the specified {@link Project}.
     * 
     * @param projectId
     *            ID of the {@link Project} to get the {@link BidRequest} for
     * @return current {@link BidRequest} resource for the {@link Project} if bids have already been
     *         requested for this {@link Project}
     * @throws CloudwordsClientException
     */
    public BidRequest getCurrentBidRequestForProject(int projectId) throws CloudwordsClientException;

    
    /**
     * Get the list of {@link CustomerBid}s for the specified Cloudwords {@link Project}.
     * 
     * @return list of {@link CustomerBid}s for the specified Cloudwords {@link Project}.
     * @throws CloudwordsClientException
     */
    public List<CustomerBid> getBids(int projectId) throws CloudwordsClientException;

    
    /**
     * Get the specified {@link CustomerBid} for the specified Cloudwords {@link Project}.
     * 
     * @return the specified {@link CustomerBid} for the specified Cloudwords {@link Project}.
     * @throws CloudwordsClientException
     */
    public CustomerBid getBid(int projectId, int bidId) throws CloudwordsClientException;
    

    /**
     * Select the winning {@link CustomerBid} for the specified Cloudwords {@link Project}.
     * 
     * @param projectId
     *            ID of the {@link Project} to select a winning bid for
     * @param bidId
     *            ID of the {@link CustomerBid} that is being selected 
     * @return current {@link BidRequest} resource for the {@link Project} 
     * @throws CloudwordsClientException
     */
    public BidRequest selectWinningBid(int projectId, int bidId) throws CloudwordsClientException;
    
    
    /**
     * Get the list of preferred {@link Vendor}s currently configured in Cloudwords.
     * 
     * @return list of preferred {@link Vendor} resources
     * @throws CloudwordsClientException
     */
    public List<Vendor> getPreferredVendors() throws CloudwordsClientException;
    
    /**
     * Get the list of preferred {@link Vendor}s currently configured in Cloudwords. This method retrieves
     * vendors that are preferred by a specific department. This version should always be used in accounts
     * with more than 0 departments.
     * 
     * @param department the department to get vendors for
     * @return list of preferred {@link Vendor} resources
     * @throws CloudwordsClientException
     */
    public List<Vendor> getPreferredVendors(Department department) throws CloudwordsClientException;

    /**
     * Get the list of source {@link Language}s currently configured in Cloudwords.
     * 
     * @return list of customer source {@link Language} resources
     * @throws CloudwordsClientException
     */
    public List<Language> getSourceLanguages() throws CloudwordsClientException;

    /**
     * Get the list of target {@link Language}s currently configured in Cloudwords.
     * 
     * @return list of customer target {@link Language} resources
     * @throws CloudwordsClientException
     */
    public List<Language> getTargetLanguages() throws CloudwordsClientException;

    /**
     * Get the list of target {@link IntendedUse}s currently configured in Cloudwords.
     * 
     * @return list of customer {@link IntendedUse} resources
     * @throws CloudwordsClientException
     */
    public List<IntendedUse> getIntendedUses() throws CloudwordsClientException;

    /**
     * Get the Cloudwords {@link Vendor} by the unique Cloudwords vendor ID.
     * 
     * @param vendorId
     *            ID of the to {@link Vendor} return
     * @return the specified {@link Vendor}
     * @throws CloudwordsClientException
     */
    public Vendor getVendor(int vendorId) throws CloudwordsClientException;

    

    /**
     * Get the specified {@link Project} {@link Task}.
     * 
     * @param projectId ID of the {@link Project} to get {@link Task} for
     * @param taskId ID of the {@link Task} to return
     * @return {@link Task} resource
     * @throws CloudwordsClientException
     */
    public CustomerTask getTask(int projectId, int taskId) throws CloudwordsClientException;
    
    /**
     * Get all {@link Task}s available to the user.
     * 
     * @return List of {@link Task}s
     * @throws CloudwordsClientException
     */
    public List<CustomerTask> getTasks() throws CloudwordsClientException;
    public List<CustomerTask> getTasks(Department department) throws CloudwordsClientException;

    /**
     * Get all {@link Task}s by status.
     * 
     * @param taskStatus The {@link Task} status to filter by
     * @return List of {@link Task}s
     * @throws CloudwordsClientException
     */
    public List<CustomerTask> getTaskByStatus(CodeAndDisplay taskStatus) throws CloudwordsClientException;
    public List<CustomerTask> getTaskByStatus(CodeAndDisplay taskStatus,Department department) throws CloudwordsClientException;

    /**
     * Get all {@link Task}s for a project.
     * 
     * @param projectId ID of the {@link Project} to get {@link Task}s for
     * @return List of {@link Task}s
     * @throws CloudwordsClientException
     */
    public List<CustomerTask> getProjectTask(int projectId) throws CloudwordsClientException;
    
    /**
     * Get all {@link Task}s for a project by status.
     * 
     * @param projectId ID of the {@link Project} to get {@link Task}s for
     * @param taskStatus The {@link Task} status to filter by
     * @return List of {@link Task}s
     * @throws CloudwordsClientException
     */
    public List<CustomerTask> getProjectTaskByStatus(int projectId, CodeAndDisplay taskStatus) throws CloudwordsClientException;
    
    /**
     * Create a {@link Task}s for a {@link Project}. 
     * 
     * @param task {@link Task} to create
     * @return Created {@link Task}
     * @throws CloudwordsClientException
     */
    public CustomerTask createTask(CustomerTask task) throws CloudwordsClientException;
    
    /**
     * Update a {@link Task} for a {@link Project}. 
     * 
     * @param task {@link Task} to update
     * @return The updated {@link Task}
     * @throws CloudwordsClientException
     */
    public CustomerTask updateTask(CustomerTask task) throws CloudwordsClientException;
    
    /**
     * Upload a {@link Task} attachment. UPLOAD CAN ONLY BE DONE ONCE, once a {@link Task} has an
     * attachment, it can not be changed.
     * 
     * @param projectId ID of the {@link Project} for the associated task
     * @param taskId ID of the {@link Task} to upload against
     * @param file
     * @return Metadata for the file uploaded
     * @throws CloudwordsClientException
     */
    public CloudwordsFile uploadTaskAttachment(int projectId, int taskId, File file) throws CloudwordsClientException;
    
    /**
     * Get all accessible {@link Department}s by the user.
     * 
     * @return List of {@link Department}s
     * @throws CloudwordsClientException
     */
    public List<Department> getDepartments() throws CloudwordsClientException;

    /**
     * Create a {@link Department} with the given name
     * @return The newly created department
     */
    public Department createDepartment(Department department) throws CloudwordsClientException;
    
    /**
     * Get the {@link User} which matches the API token in use.
     * @return the current user
     * @throws CloudwordsClientException
     */
    public User getCurrentUser() throws CloudwordsClientException;
    
    /**
     * Get a list of all {@link User}s in all {@link Department}s the user is a member of. If departments are not enabled,
     * all users are returned. 
     * @return a list of users
     * @throws CloudwordsClientException
     */
    public List<User> getUsers() throws CloudwordsClientException;
    
    /**
     * Get a list of {@link User}s in the specified {@link Department}.
     * @return all members of the specified department
     * @throws CloudwordsClientException
     */
    public List<User> getUsers(Department department) throws CloudwordsClientException;
    
    /**
     * For accounts which do not have departments enabled, this returns a list of users available to follow projects.
     * For accounts which do have departments enabled, this returns a list of users available to follow projects in the Unassigned department.
     * @return A list of users which may follow projects.
     * @throws CloudwordsClientException
     */
    public List<User> getAvailableFollowers() throws CloudwordsClientException;
    
    /**
     * This retrieves a list of users who are available to follow projects in the specified {@link Department}.
     */
    public List<User> getAvailableFollowers(Department department) throws CloudwordsClientException;
    
    /**
     * Get the specified {@link ReviewerInstructions} metadata. 
     * 
     * @param projectId
     *            ID of the {@link Project} to get {@link ReviewerInstructions} for
     * @param languageCode
     *            {@link Language} code of the {@link ReviewerInstructions} to return
     * @return {@link ReviewerInstructions} resource
     * @throws CloudwordsClientException
     */
    public ReviewerInstructions getReviewerInstructions(int projectId, String languageCode) throws CloudwordsClientException;

    /**
     * Get the specified {@link ReviewerInstructions} metadata. 
     * 
     * @param riId
     *            ID of the {@link ReviewerInstructions} to get
     * @return {@link ReviewerInstructions} resource
     * @throws CloudwordsClientException
     */
    public ReviewerInstructions getReviewerInstructions(int riId) throws CloudwordsClientException;
    
    /**
     * Add the specified {@link ReviewerInstructions}. 
     * 
     * @param projectId
     *            ID of the {@link Project} to add {@link ReviewerInstructions} for
     * @param languageCode
     *            {@link Language} code to add {@link ReviewerInstructions} for
     * @param reviewerInstructions
     *            {@link ReviewerInstructions} to add           
     * @return {@link ReviewerInstructions} created resource
     * @throws CloudwordsClientException
     */
    public ReviewerInstructions createReviewerInstructions(int projectId, String languageCode, ReviewerInstructions reviewerInstructions) throws CloudwordsClientException;
    
    /**
     * Update the specified {@link ReviewerInstructions}. 
     * 
     * @param reviewerInstructions
     *            {@link ReviewerInstructions} to update   
     * @return {@link ReviewerInstructions} updated resource
     * @throws CloudwordsClientException
     */
    public ReviewerInstructions updateReviewerInstructions(ReviewerInstructions reviewerInstructions) throws CloudwordsClientException;
    
    /**
     * Delete the specified {@link ReviewerInstructions}. 
     * 
     * @param riId
     *            ID of the {@link ReviewerInstructions} to get 
     * @throws CloudwordsClientException
     */
    public void deleteReviewerInstructions (int riId) throws CloudwordsClientException;

    /**
     * Creates a new Cloudwords {@link CustomerChangeOrder} within a {@link Project}. Please refer to the API
     * documentation for more information.
     * 
     * @param projectId
     *            ID of the {@link Project} to add {@link CustomerChangeOrder} for
     * @param customerChangeOrder
     *            {@link CustomerChangeOrder} to add
     * @return {@link CustomerChangeOrder} created resource
     * @throws CloudwordsClientException
     */
    public CustomerChangeOrder createChangeOrder(int projectId, CustomerChangeOrder customerChangeOrder)
            throws CloudwordsClientException;

    /**
     * Gets the Cloudwords {@link CustomerChangeOrder} associated with the unique Cloudwords {@link CustomerChangeOrder}
     * ID. Please refer to the API documentation for more information.
     * @param projectId
     *              ID of the {@link Project} of the associated {@link CustomerChangeOrder}
     * @param changeOrderId
     *            ID of the {@link CustomerChangeOrder} to return
     * @return {@link CustomerChangeOrder} resource
     * @throws CloudwordsClientException
     */
    public CustomerChangeOrder getChangeOrder(int projectId, int changeOrderId) throws CloudwordsClientException;

    /**
     * Gets all {@link CustomerChangeOrder}s for a project. Please refer to the API documentation for more information.
     * 
     * @param projectId
     *            ID of the {@link Project} to get {@link CustomerChangeOrder}s for
     * @return List of {@link CustomerChangeOrder}s
     * @throws CloudwordsClientException
     */
    public List<CustomerChangeOrder> getChangeOrders(int projectId) throws CloudwordsClientException;

    /**
     * Updates the specified {@link CustomerChangeOrder} by canceling the
     * {@link CustomerChangeOrder}. Please refer to the API documentation for more information.
     * 
     * @param customerChangeOrder
     *            {@link CustomerChangeOrder} to update
     * @return {@link CustomerChangeOrder} updated resource
     * @throws CloudwordsClientException
     */
    public CustomerChangeOrder cancelChangeOrder(CustomerChangeOrder customerChangeOrder)
            throws CloudwordsClientException;
    
    /**
     * Updates the specified {@link CustomerChangeOrder} by submitting the
     * {@link CustomerChangeOrder} to the vendor. Please refer to the API documentation for more information.
     * 
     * @param customerChangeOrder
     *            {@link CustomerChangeOrder} to update
     * @return {@link CustomerChangeOrder} updated resource
     * @throws CloudwordsClientException
     */
    public CustomerChangeOrder submitChangeOrder(CustomerChangeOrder customerChangeOrder)
            throws CloudwordsClientException;
    
    /**
     * Updates the specified {@link CustomerChangeOrder} by accepting the {@link CustomerChangeOrderBid} submitted by
     * the vendor. The project state will be changed to include any new languages, and the document of the primary
     * project will have a new version using the files of the change order. Please refer to the API documentation for
     * more information.
     * 
     * @param customerChangeOrder
     *            {@link CustomerChangeOrder} to update
     * @return {@link CustomerChangeOrder} updated resource
     * @throws CloudwordsClientException
     */
    public CustomerChangeOrder acceptChangeOrder(CustomerChangeOrder customerChangeOrder)
            throws CloudwordsClientException;

    

    /**
     * Uploads the source material for a given Cloudwords {@link CustomerChangeOrder}.
     * <p>
     * Please note that this {@link CustomerChangeOrder} must already exists in Cloudwords, and the
     * {@link CustomerChangeOrder} must be in a state where source material can be uploaded. If source material for this
     * {@link CustomerChangeOrder} already exists, then this operation will overwrite the source material associated
     * with the {@link CustomerChangeOrder}. Please refer to the API documentation for more information.
     * </p>
     * @param projectId
     *              ID of the {@link Project} of the associated {@link CustomerChangeOrder}
     * @param changeOrderId
     *            ID of {@link CustomerChangeOrder} to upload source for
     * @param file
     *            A single file to be uploaded for this {@link CustomerChangeOrder}
     * @return Uploaded source material {@link CloudwordsFile} resource metadata
     * @throws CloudwordsClientException
     */
    public CloudwordsFile uploadChangeOrderFile(int projectId, int changeOrderId, File file)
            throws CloudwordsClientException;

    /**
     * Gets the specified {@link CustomerChangeOrderBid} submitted by the vendor for the specified Cloudwords
     * {@link CustomerChangeOrder}. Please refer to the API documentation for more information.
     * @param projectId
     *              ID of the {@link Project} of the associated {@link CustomerChangeOrder}
     * @param changeOrderId
     *            ID of {@link CustomerChangeOrder} to get the bid for
     * @return the specified {@link CustomerChangeOrderBid} for the specified Cloudwords {@link CustomerChangeOrder}.
     * @throws CloudwordsClientException
     */
    public CustomerChangeOrderBid getChangeOrderBid(int projectId, int changeOrderId) throws CloudwordsClientException;
    
    /**
     * Update a project workflow, this allows the phase due dates and assignees to be changed. Restrictions for these
     * fields can be found in the API documentation.
     * @param projectId The project ID that the workflow belongs to
     * @param workflow The workflow to update
     * @return The updated workflow
     * @throws CloudwordsClientException
     */
    public ProjectWorkflow updateProjectWorkflow(int projectId, ProjectWorkflow workflow) throws CloudwordsClientException;
    
    /**
     * Change the template workflow to use for a project. This can only be called before bid-out, after target languages
     * and delivery due date have been set on the project.
     * @param projectId The ID of the project
     * @param templateId The ID of the template (derived from getAvailableWorkflowTemplates)
     * @throws CloudwordsClientException
     */
    public List<ProjectWorkflow> selectProjectWorkflow(int projectId, int templateId) throws CloudwordsClientException;
    
    /**
     * Get the list of available workflow templates for the project. This method may return options even if a workflow
     * cannot be selected on the project at present time.
     * @param projectId The ID of the project
     * @throws CloudwordsClientException
     */
    public List<WorkflowTemplate> getAvailableWorkflowTemplates(int projectId) throws CloudwordsClientException;

    /**
     * Get a specific template that is available on a project by ID.
     * @param projectId the ID of the project
     * @param templateId The ID of the template to fetch 
     * @throws CloudwordsClientException
     */
    public WorkflowTemplate getWorkflowTemplate(int projectId, int templateId) throws CloudwordsClientException;
    
    /**
     * Update a change order project workflow, this allows the phase due dates and assignees to be changed. Restrictions for these
     * fields can be found in the API documentation.
     * @param projectId The ID of the project
     * @param changeOrderId The ID of the change order
     * @param workflow The workflow to update
     */
    public ChangeOrderProjectWorkflow updateChangeOrderProjectWorkflow(int projectId, int changeOrderId, ChangeOrderProjectWorkflow workflow) throws CloudwordsClientException;

    /**
     * <p>
     * Upload a new translated document preview. The translated document ID should be derived from {@link TranslatedDocument#getId()}. 
     * </p>
     * @param projectId ID of the {@link Project} to upload the translated document preview for
     * @param language The specific {@link Language}
     * @param translatedDocumentId The ID of the translated document
     * @param file The translation preview to upload. This should be a zip file containing a single html file and all referenced assets (images, css, js). All paths should be relative.
     * @throws CloudwordsClientException
     */
    public void addTranslatedDocumentPreview(int projectId, Language language, Integer translatedDocumentId, File file) throws CloudwordsClientException;
    
    /**
     * <p>
     * Upload a new source document preview. The source document ID should be derived from {@link SourceDocument#getId()}. 
     * </p>
     * @param projectId ID of the {@link Project} to upload the translated document preview for
     * @param sourceDocumentId The ID of the source document
     * @param file The source preview to upload. This should be a zip file containing a single html file and all referenced assets (images, css, js). All paths should be relative.
     * @throws CloudwordsClientException
     */
    public void addSourceDocumentPreview(int projectId, Integer sourceDocumentId, File file) throws CloudwordsClientException;
    
    /**
     * For accounts which do not have departments enabled, this returns a list of users which can be assoicated as a project owner
     * For accounts which do have departments enabled, this returns a list of users available as a project owner in the Unassigned department.
     * @return A list of users which may be a project owner
     * @throws CloudwordsClientException
     */
    public List<User> getAvailableProjectOwners() throws CloudwordsClientException;
    
    /**
     * This retrieves a list of users who are available to follow projects in the specified {@link Department}.
     */
    public List<User> getAvailableProjectOwners(Department department) throws CloudwordsClientException;
    
    /**
     * Get the current {@link TranslationRequest} for the specified {@link Project}.
     * 
     * @param projectId
     *            ID of the {@link Project} to get the {@link TranslationRequest} for
     * @return current {@link TranslationRequest} resource for the {@link Project} if requests have already been
     *         requested for this {@link Project}
     * @throws CloudwordsClientException
     */
    public CustomerTranslationRequest getCurrentTranslationRequestForProject(int projectId) throws CloudwordsClientException;
    
    /**
     * Request translation request for the specified {@link Project}. In order to request a translation for a
     * {@link Project}, a source material must be already uploaded for the {@link Project}, and the
     * {@link Project} must be in a state where requesting bids is allowed. A translation request must be
     * sent to a preferred {@link Vendor} configured to recieve bidless translation requests.
     * 
     * Please refer to the API documentation for more information.
     * 
     * @param projectId
     *            ID of the {@link Project} to request translation for
     * @param transRequest
     *            A {@link CustomerTranslationRequest}s to send
     * @throws CloudwordsClientException
     */
    public CustomerTranslationRequest requestTranslationRequestForProject(int projectId, CustomerTranslationRequest transRequest) 
            throws CloudwordsClientException;
    
    /**
     * Select the winning {@link CustomerBid} for the specified Cloudwords {@link Project}.
     * 
     * @param projectId
     *            ID of the {@link Project} to select a winning bid for
     * @param bidId
     *            ID of the {@link CustomerBid} that is being selected 
     * @return current {@link CustomerTranslationRequest} resource for the {@link Project} 
     * @throws CloudwordsClientException
     */
    public CustomerTranslationRequest selectWinningTranslationRequest(int projectId, int bidId) throws CloudwordsClientException;

    /**
     * Search for projects in a department and return them in a paged format. See API documentation for details.
     * 
     * @since 1.21
     * @throws CloudwordsClientException
     */
    public PagedResponse<Project> getProjects(Department department, PagedProjectQuery query) throws CloudwordsClientException;
}
