/*
 * Copyright 2012, Cloudwords, Inc.
 *
 * Licensed under the API LICENSE AGREEMENT, Version 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.cloudwords.com/developers/license-1.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.cloudwords.api.client;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.http.HttpEntity;
import org.apache.http.HttpHost;
import org.apache.http.client.CredentialsProvider;
import org.apache.http.client.HttpClient;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.client.methods.HttpDelete;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.methods.HttpPut;
import org.apache.http.client.methods.HttpUriRequest;
import org.apache.http.config.SocketConfig;
import org.apache.http.entity.ContentType;
import org.apache.http.entity.StringEntity;
import org.apache.http.entity.mime.MultipartEntityBuilder;
import org.apache.http.entity.mime.content.FileBody;
import org.apache.http.entity.mime.content.StringBody;
import org.apache.http.impl.client.HttpClientBuilder;

import com.cloudwords.api.client.exception.CloudwordsClientException;
import com.cloudwords.api.client.resources.BidOptions;
import com.cloudwords.api.client.resources.BidRequest;
import com.cloudwords.api.client.resources.ChangeOrderProjectWorkflow;
import com.cloudwords.api.client.resources.CloudwordsFile;
import com.cloudwords.api.client.resources.CloudwordsLanguageFile;
import com.cloudwords.api.client.resources.CodeAndDisplay;
import com.cloudwords.api.client.resources.CustomerBid;
import com.cloudwords.api.client.resources.CustomerChangeOrder;
import com.cloudwords.api.client.resources.CustomerChangeOrderBid;
import com.cloudwords.api.client.resources.CustomerTask;
import com.cloudwords.api.client.resources.CustomerTranslationRequest;
import com.cloudwords.api.client.resources.Department;
import com.cloudwords.api.client.resources.IntendedUse;
import com.cloudwords.api.client.resources.Language;
import com.cloudwords.api.client.resources.PagedProjectQuery;
import com.cloudwords.api.client.resources.PagedResponse;
import com.cloudwords.api.client.resources.Project;
import com.cloudwords.api.client.resources.ProjectWorkflow;
import com.cloudwords.api.client.resources.ReviewerInstructions;
import com.cloudwords.api.client.resources.SourceDocument;
import com.cloudwords.api.client.resources.TranslatedDocument;
import com.cloudwords.api.client.resources.User;
import com.cloudwords.api.client.resources.Vendor;
import com.cloudwords.api.client.resources.Webhook;
import com.cloudwords.api.client.resources.WorkflowTemplate;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;

/**
 * Basic implementation of the Cloudwords Customer API client.
 * 
 * @author jonathansu
 * @since 1.3
 */
public class CloudwordsCustomerClient extends BaseCloudwordsClient implements CloudwordsCustomerAPI {

    // The authorization token to validate identify when accessing the Cloudwords API
    private final String authToken;
    // The domain + version for connecting to the API (e.g. https://api.cloudwords.com/v1)
    private final ObjectMapper mapper = new ObjectMapper();
    // The main HTTP client that will send/receive communication with the Cloudwords API
    private HttpClient client;
    
    /**
     * Convenience constructor that provides default configuration.
     * 
     * @param baseAPIURL
     *            The base domain of the Cloudwords API
     * @param version
     *            The version of the Cloudwords API to use
     * @param authorizationToken
     *            The authorization token to validate identify when accessing the Cloudwords API
     */
    public CloudwordsCustomerClient(String baseAPIURL, String version, String authorizationToken) {
        this(baseAPIURL, version, authorizationToken, connectionTimeoutMillis, socketTimeoutMillis, maxTotalConnections);
    }
    
    /**
     * Convenience constructor that provides default configuration.
     * 
     * @param baseAPIURL
     *            The base domain of the Cloudwords API
     * @param version
     *            The version of the Cloudwords API to use
     * @param authorizationToken
     *            The authorization token to validate identify when accessing the Cloudwords API
     * @param proxy
     *            The proxy server to use for connections
     * @param proxyCredentialsProvider
     *            The credentials provider to use with the proxy server
     */
    public CloudwordsCustomerClient(String baseAPIURL, String version, String authorizationToken, HttpHost proxy, CredentialsProvider proxyCredentialsProvider) {
        this(baseAPIURL, version, authorizationToken, connectionTimeoutMillis, socketTimeoutMillis, maxTotalConnections, proxy, proxyCredentialsProvider);
    }
    
    /**
     * Base constructor.
     * 
     * @param baseAPIURL
     *            The base domain of the Cloudwords API
     * @param version
     *            The version of the Cloudwords API to use
     * @param authorizationToken
     *            The authorization token to validate identify when accessing the Cloudwords API
     * @param connectionTimeoutMillis
     *            Determines the timeout in milliseconds until a connection is established. A
     *            timeout value of zero is interpreted as an infinite timeout.
     * @param socketTimeoutMillis
     *            Defines the socket timeout (<code>SO_TIMEOUT</code>) in milliseconds, which is the
     *            timeout for waiting for data or, put differently, a maximum period inactivity
     *            between two consecutive data packets). A timeout value of zero is interpreted as
     *            an infinite timeout.
     * @param maxTotalConnections
     *            The max concurrent connections the client can establish against the Cloudwords API
     * @param proxy
     *            The proxy server to use for connections
     * @param proxyCredentialsProvider
     *            The credentials provider to use with the proxy server
     */
    public CloudwordsCustomerClient(String baseAPIURL, String version, String authorizationToken, int connectionTimeoutMillis,
            int socketTimeoutMillis, int maxTotalConnections, HttpHost proxy, CredentialsProvider proxyCredentialsProvider) {
        this(baseAPIURL, version, authorizationToken, null);
        
        if (maxTotalConnections > 5) {
            throw new IllegalArgumentException("Only 5 concurrent connections per Cloudwords account is supported");
        }

        HttpClientBuilder clientBuilder = HttpClientBuilder.create()
            .setDefaultSocketConfig(SocketConfig.custom().setSoTimeout(socketTimeoutMillis).build())
            .setDefaultRequestConfig(RequestConfig.custom().setConnectTimeout(connectionTimeoutMillis).setSocketTimeout(socketTimeoutMillis).build())
            .setMaxConnTotal(maxTotalConnections);
        
        if (proxy != null) {
            clientBuilder.setProxy(proxy);
            if (proxyCredentialsProvider != null) {
                clientBuilder.setDefaultCredentialsProvider(proxyCredentialsProvider);
            }
        }
        
        this.client = clientBuilder.build();
    }
    
    /**
     * Creates a new client that takes a user-provided instance of HttpClient. The caller is responsible for ensuring this client is properly initialized.
     * As of HttpClient 4.3.5, the following code is used by the other constructors in this class to build a proper client:
     * <p>
     * HttpClientBuilder.create()
            .setDefaultSocketConfig(SocketConfig.custom().setSoTimeout(socketTimeoutMillis).build())
            .setDefaultRequestConfig(RequestConfig.custom().setConnectTimeout(connectionTimeoutMillis).setSocketTimeout(socketTimeoutMillis).build())
            .setMaxConnTotal(maxTotalConnections).build();
        </p>
     * 
     * @param baseAPIURL The base domain of the Cloudwords API
     * @param version The version of the Cloudwords API to use
     * @param authorizationToken The authorization token to validate identify when accessing the Cloudwords API
     * @param client The HTTP Client instance to use
     */
    public CloudwordsCustomerClient(String baseAPIURL, String version, String authorizationToken, HttpClient client)    {
        this.apiVersion = Double.parseDouble(version);
        this.authToken = new StringBuilder("UserToken ").append(authorizationToken).toString();
        this.baseUrlWithVersion = baseAPIURL + "/" + version;
        this.client = client;
        initObjectMapper();
    }
    

    /**
     * Base constructor.
     * 
     * @param baseAPIURL
     *            The base domain of the Cloudwords API
     * @param version
     *            The version of the Cloudwords API to use
     * @param authorizationToken
     *            The authorization token to validate identify when accessing the Cloudwords API
     * @param connectionTimeoutMillis
     *            Determines the timeout in milliseconds until a connection is established. A
     *            timeout value of zero is interpreted as an infinite timeout.
     * @param socketTimeoutMillis
     *            Defines the socket timeout (<code>SO_TIMEOUT</code>) in milliseconds, which is the
     *            timeout for waiting for data or, put differently, a maximum period inactivity
     *            between two consecutive data packets). A timeout value of zero is interpreted as
     *            an infinite timeout.
     * @param maxTotalConnections
     *            The max concurrent connections the client can establish against the Cloudwords API
     */
    public CloudwordsCustomerClient(String baseAPIURL, String version, String authorizationToken, int connectionTimeoutMillis,
            int socketTimeoutMillis, int maxTotalConnections) {
        this(baseAPIURL, version, authorizationToken, connectionTimeoutMillis, socketTimeoutMillis, maxTotalConnections, null, null);
    }

    @Override
    public Project getProject(int projectId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + ".json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        return executeRequest(request, Project.class);
    }

    @Override
    public Project createProject(Project project) throws CloudwordsClientException {
        if (project != null) {
            HttpPost request = new HttpPost(baseUrlWithVersion + "/project");
            setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

            try {
                request.setEntity(new StringEntity(mapper.writeValueAsString(project),UTF_8));
            } catch (Exception ex) {
                throw new CloudwordsClientException(ex);
            }
            return executeRequest(request, Project.class);
        }
        return null;
    }
    
    @Override
    public CloudwordsFile uploadProjectSource(int projectId, File zipFile) throws CloudwordsClientException {
        if (zipFile != null) {
            validateZipFile(zipFile);
            return putFile("/project/" + projectId + "/file/source", zipFile, CloudwordsFile.class);
		} else {
			throw new CloudwordsClientException(
					"Unable to upload project source because no file was specified. Project id: '"
							+ projectId + "'");
		}
    }
    
    @Override
    public BidRequest requestBidsForProject(int projectId, List<Vendor> preferredVendors, boolean doLetCloudwordsChoose, boolean doAutoSelectBidFromVendor)
            throws CloudwordsClientException {

        if ((preferredVendors == null || preferredVendors.size() == 0) && !doLetCloudwordsChoose) {
            throw new CloudwordsClientException(
                    "Unable to request bids for project because you must either specify"
                            + " the Cloudwords preferred vendors to request bids from or allow Cloudwords to automatically find"
                            + " vendors that best match the project requirements, or both. Project ID: '" + projectId
                            + "'");
        }
        if (doAutoSelectBidFromVendor && (preferredVendors == null || preferredVendors.size() != 1 || doLetCloudwordsChoose)) {
            throw new CloudwordsClientException(
                    "Unable to request bids for project because"
                            +  " it is not eligible for bid auto-selection.  To be eligible your bid request must have exactly one preferred provider selected and the doLetCloudwordsChoose option disabled");
        }
        HttpPost request = new HttpPost(baseUrlWithVersion + "/project/" + projectId + "/bid-request");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        BidRequest br = new BidRequest();
        br.setPreferredVendors(preferredVendors);
        br.setDoLetCloudwordsChoose(doLetCloudwordsChoose);
        br.setDoAutoSelectBidFromVendor(doAutoSelectBidFromVendor);
        try {
            request.setEntity(new StringEntity(mapper.writeValueAsString(br),UTF_8));
        } catch (Exception ex) {
            throw new CloudwordsClientException(ex);
        }
        return executeRequest(request, BidRequest.class);
    }

    @Override
    public BidRequest getCurrentBidRequestForProject(int projectId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/bid-request/current.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        return executeRequest(request, BidRequest.class);
    }
    
    @Override
    public List<CustomerBid> getBids(int projectId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/bid.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        
        TypeReference<List<CustomerBid>> typeRef = new TypeReference<List<CustomerBid>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public CustomerBid getBid(int projectId, int bidId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/bid/" + bidId + ".json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        return executeRequest(request, CustomerBid.class);
    }
    
    @Override
    public BidRequest selectWinningBid(int projectId, int bidId) throws CloudwordsClientException {
        HttpPut request = new HttpPut(baseUrlWithVersion + "/project/" + projectId + "/bid-request");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        try {
            BidRequest bidRequest = new BidRequest();
            bidRequest.setWinningBidId(bidId);
            request.setEntity(new StringEntity(mapper.writeValueAsString(bidRequest),UTF_8));
        } catch (Exception ex) {
            throw new CloudwordsClientException(ex);
        }
        return executeRequest(request, BidRequest.class);
    }
    
    @Override
    public InputStream downloadSourceFile(int projectId) throws CloudwordsClientException {
        InputStream stream = null;
        try {
            CloudwordsFile metadata = getProjectSource(projectId);
            stream = downloadFileFromMetadata(metadata);
            return stream;
        } catch (Exception ex) {
            try {
            	if(stream != null) {
                    stream.close();            		
            	}
            } catch (IOException ioe) {
                throw new CloudwordsClientException(ioe);
            }
            throw new CloudwordsClientException(ex);
        }
    }

    @Override
    public List<Vendor> getPreferredVendors() throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/vendor/preferred.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        TypeReference<List<Vendor>> typeRef = new TypeReference<List<Vendor>>() {};
        return executeRequest(request, typeRef);
    }
    
    @Override
    public List<Vendor> getPreferredVendors(Department department) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/department/" + department.getId() + "/vendor/preferred.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        TypeReference<List<Vendor>> typeRef = new TypeReference<List<Vendor>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public List<Language> getSourceLanguages() throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/org/settings/project/language/source.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        TypeReference<List<Language>> typeRef = new TypeReference<List<Language>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public List<Language> getTargetLanguages() throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/org/settings/project/language/target.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        TypeReference<List<Language>> typeRef = new TypeReference<List<Language>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public List<IntendedUse> getIntendedUses() throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/org/settings/project/intended-use.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        TypeReference<List<IntendedUse>> typeRef = new TypeReference<List<IntendedUse>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public Vendor getVendor(int vendorId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/vendor/" + vendorId + ".json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        return executeRequest(request, Vendor.class);
    }

    @Override
    public List<Project> getOpenProjects() throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/open.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        TypeReference<List<Project>> typeRef = new TypeReference<List<Project>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public List<Project> getOpenProjects(Department department) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/department/" + department.getId() + "/project/open.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        TypeReference<List<Project>> typeRef = new TypeReference<List<Project>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public List<Project> getClosedProjects() throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/closed.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        TypeReference<List<Project>> typeRef = new TypeReference<List<Project>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public List<Project> getClosedProjects(Department department) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/department/" + department.getId() + "/project/closed.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        TypeReference<List<Project>> typeRef = new TypeReference<List<Project>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public Project updateProject(Project project) throws CloudwordsClientException {
        if (project != null) {
            HttpPut request = new HttpPut(baseUrlWithVersion + "/project/" + project.getId());
            setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

            try {
                request.setEntity(new StringEntity(mapper.writeValueAsString(project),UTF_8));
            } catch (Exception ex) {
                throw new CloudwordsClientException(ex);
            }
            return executeRequest(request, Project.class);
        }
        return null;
    }

    @Override
    public CloudwordsFile getProjectSource(int projectId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/file/source.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        return executeRequest(request, CloudwordsFile.class);
    }
    
    @Override
    public CloudwordsFile uploadProjectReference(int projectId, File zipFile) throws CloudwordsClientException {
		if (zipFile != null) {
		    validateZipFile(zipFile);
		    return postFile("/project/" + projectId + "/file/reference", zipFile, CloudwordsFile.class);
		} else {
			throw new CloudwordsClientException(
					"Unable to upload project reference because no file was specified. Project id: '"
							+ projectId + "'");
		}
    }
    
    
    @Override
    public CloudwordsFile updateProjectReference(int projectId, int documentId, File zipFile)
            throws CloudwordsClientException {
		if (zipFile != null) {
		    validateZipFile(zipFile);
		    return putFile("/project/" + projectId + "/file/reference/" + documentId, zipFile, CloudwordsFile.class);
		} else {
			throw new CloudwordsClientException(
					"Unable to upload project reference because no file was specified. Project id: '"
							+ projectId + "'");
		}
    }

    @Override
    public List<CloudwordsFile> getProjectReferences(int projectId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/file/reference.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        TypeReference<List<CloudwordsFile>> typeRef = new TypeReference<List<CloudwordsFile>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public CloudwordsFile getProjectReference(int projectId, int documentId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/file/reference/"
                + documentId + ".json");

        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        return executeRequest(request, CloudwordsFile.class);
    }

    @Override
    public InputStream downloadReferenceFile(int projectId, int documentId) throws CloudwordsClientException {
        InputStream stream = null;
        try {
            CloudwordsFile metadata = getProjectReference(projectId, documentId);
            stream = downloadFileFromMetadata(metadata);
            return stream;
        } catch (Exception ex) {
            try {
            	if(stream != null) {
                    stream.close();            		
            	}
            } catch (IOException ioe) {
                throw new CloudwordsClientException(ioe);
            }
            throw new CloudwordsClientException(ex);
        }
    }

    @Override
    public CloudwordsFile getMasterProjectTranslatedFile(int projectId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/file/translated.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        return executeRequest(request, CloudwordsFile.class);
    }

    @Override
    public List<CloudwordsLanguageFile> getProjectTranslatedFiles(int projectId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId
                + "/file/translated/language.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        TypeReference<List<CloudwordsLanguageFile>> typeRef = new TypeReference<List<CloudwordsLanguageFile>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public CloudwordsLanguageFile getProjectTranslatedFile(int projectId, Language language)
            throws CloudwordsClientException {
        if (language != null && language.getLanguageCode() != null) {
            HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId
                    + "/file/translated/language/" + language.getLanguageCode() + ".json");

            setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
            return executeRequest(request, CloudwordsLanguageFile.class);
        } else {
            throw new CloudwordsClientException(
                    "Unable to get project translated file for language because either the language object or the language code is null");
        }
    }

    @Override
    public InputStream downloadMasterTranslatedFile(int projectId) throws CloudwordsClientException {
        InputStream stream = null;
        try {
            CloudwordsFile metadata = getMasterProjectTranslatedFile(projectId);
            stream = downloadFileFromMetadata(metadata);
            return stream;
        } catch (Exception ex) {
            try {
            	if(stream != null) {
                    stream.close();            		
            	}
            } catch (IOException ioe) {
                throw new CloudwordsClientException(ioe);
            }
            throw new CloudwordsClientException(ex);
        }
    }

    @Override
    public InputStream downloadTranslatedFile(int projectId, Language language) throws CloudwordsClientException {
        InputStream stream = null;
        try {
            CloudwordsLanguageFile metadata = getProjectTranslatedFile(projectId, language);
            stream = downloadFileFromMetadata(metadata);
            return stream;
        } catch (Exception ex) {
            try {
            	if(stream != null) {
                    stream.close();            		
            	}
            } catch (IOException ioe) {
                throw new CloudwordsClientException(ioe);
            }
            throw new CloudwordsClientException(ex);
        }
    }

	@Override
    public CloudwordsLanguageFile approveTranslatedFile(int projectId, Language language) throws CloudwordsClientException {
	    String url =  "/project/" + projectId + "/file/translated/language/" + language.getLanguageCode();
	    
	    if (apiVersion >= 1.14)    {
	        url += ".json";
	    }
	    
	    CloudwordsLanguageFile languageFile = new CloudwordsLanguageFile();
        CodeAndDisplay langStatus = new CodeAndDisplay();
        langStatus.setCode("approved");
        languageFile.setStatus(langStatus);
        
        return put(url, languageFile);
    }

    @Override
    public CloudwordsFile getProjectTranslationMemory(int projectId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/file/tm.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        return executeRequest(request, CloudwordsFile.class);
    }
    
    @Override
    public InputStream downloadProjectTranslationMemory(int projectId) throws CloudwordsClientException {
        InputStream stream = null;
        try {
            CloudwordsFile metadata = getProjectTranslationMemory(projectId);
            stream = downloadFileFromMetadata(metadata);
            return stream;
        } catch (Exception ex) {
            try {
                if(stream != null) {
                    stream.close();                 
                }
            } catch (IOException ioe) {
                throw new CloudwordsClientException(ioe);
            }
            throw new CloudwordsClientException(ex);
        }
    }

    @Override
    protected HttpClient getClient() {
        return this.client;
    }

    @Override
    protected ObjectMapper getMapper() {
        return this.mapper;
    }

    @Override
    protected String getAuthToken() {
        return this.authToken;
    }

    @Override
    public CustomerTask getTask(int projectId, int taskId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/task/" + taskId + ".json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        TypeReference<CustomerTask> typeRef = new TypeReference<CustomerTask>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public List<CustomerTask> getTasks() throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/task.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        TypeReference<List<CustomerTask>> typeRef = new TypeReference<List<CustomerTask>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public List<CustomerTask> getTasks(Department department) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/department/" + department.getId() + "/task.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        TypeReference<List<CustomerTask>> typeRef = new TypeReference<List<CustomerTask>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public List<CustomerTask> getTaskByStatus(CodeAndDisplay taskStatus) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/task/status/" + taskStatus.getCode() + ".json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        TypeReference<List<CustomerTask>> typeRef = new TypeReference<List<CustomerTask>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public List<CustomerTask> getTaskByStatus(CodeAndDisplay taskStatus,Department department) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/department/" + department.getId() + "/task/status/" + taskStatus.getCode() + ".json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        TypeReference<List<CustomerTask>> typeRef = new TypeReference<List<CustomerTask>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public List<CustomerTask> getProjectTask(int projectId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/task.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        TypeReference<List<CustomerTask>> typeRef = new TypeReference<List<CustomerTask>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public List<CustomerTask> getProjectTaskByStatus(int projectId, CodeAndDisplay taskStatus) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/task/status/"
                + taskStatus.getCode() + ".json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        TypeReference<List<CustomerTask>> typeRef = new TypeReference<List<CustomerTask>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public CustomerTask createTask(CustomerTask task) throws CloudwordsClientException {
        if (task != null) {
            HttpPost request = new HttpPost(baseUrlWithVersion + "/project/" + task.getProject().getId() + "/task.json");
            setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
            try {
                request.setEntity(new StringEntity(mapper.writeValueAsString(task),UTF_8));
            } catch (Exception ex) {
                throw new CloudwordsClientException(ex);
            }
            return executeRequest(request, CustomerTask.class);
        }
        return null;
    }

    @Override
    public CustomerTask updateTask(CustomerTask task) throws CloudwordsClientException {
        if (task != null) {
            HttpPut request = new HttpPut(baseUrlWithVersion + "/project/" + task.getProject().getId() + "/task/"
                    + task.getId() + ".json");
            setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
            try {
                request.setEntity(new StringEntity(mapper.writeValueAsString(task),UTF_8));
            } catch (Exception ex) {
                throw new CloudwordsClientException(ex);
            }
            return executeRequest(request, CustomerTask.class);
        }
        return null;
    }

    @Override
    public CloudwordsFile getTaskAttachment(int projectId, int taskId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/"+projectId+"/task/" + taskId
                + "/file/attachment.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        return executeRequest(request, CloudwordsFile.class);
    }

    @Override
    public InputStream downloadTaskAttachment(int projectId, int taskId) throws CloudwordsClientException {
        InputStream stream = null;
        try {
            CloudwordsFile metadata = getTaskAttachment(projectId, taskId);
            stream = downloadFileFromMetadata(metadata);
            return stream;
        } catch (Exception ex) {
            try {
                if (stream != null) {
                    stream.close();
                }
            } catch (IOException ioe) {
                throw new CloudwordsClientException(ioe);
            }
            throw new CloudwordsClientException(ex);
        }
    }

    @Override
    public CloudwordsFile uploadTaskAttachment(int projectId, int taskId, File file) throws CloudwordsClientException {
        if (file != null) {
            return putFile("/project/"+projectId+"/task/" + taskId
                    + "/file/attachment/content", file, CloudwordsFile.class);
        } else {
            throw new CloudwordsClientException("Unable to upload because no file was specified");
        }
    }

    @Override
    public List<Department> getDepartments() throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/department.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        TypeReference<List<Department>> typeRef = new TypeReference<List<Department>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public Department createDepartment(Department department) throws CloudwordsClientException {
        if (department != null) {
            HttpPost request = new HttpPost(baseUrlWithVersion + "/department.json");
            setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
            try {
                request.setEntity(new StringEntity(mapper.writeValueAsString(department),UTF_8));
            } catch (Exception ex) {
                throw new CloudwordsClientException(ex);
            }
            return executeRequest(request, Department.class);
        }
        return null;
    }

    @Override
    public User getCurrentUser() throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/user/current.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        TypeReference<User> typeRef = new TypeReference<User>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public List<User> getUsers() throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/user.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        TypeReference<List<User>> typeRef = new TypeReference<List<User>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public List<User> getUsers(Department department) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/department/" + department.getId() + "/user.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        TypeReference<List<User>> typeRef = new TypeReference<List<User>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public List<User> getAvailableFollowers() throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/follower/available.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        TypeReference<List<User>> typeRef = new TypeReference<List<User>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public List<User> getAvailableFollowers(Department department) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/department/" + department.getId() + "/follower/available.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        TypeReference<List<User>> typeRef = new TypeReference<List<User>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public ReviewerInstructions getReviewerInstructions(int projectId, String languageCode)
            throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/language/" + languageCode + "/reviewer-instructions.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        return executeRequest(request, ReviewerInstructions.class);
    }

    @Override
    public ReviewerInstructions getReviewerInstructions(int riId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/reviewer-instructions/" + riId + ".json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        return executeRequest(request, ReviewerInstructions.class);
    }

    @Override
    public ReviewerInstructions createReviewerInstructions(int projectId, String languageCode,
            ReviewerInstructions reviewerInstructions) throws CloudwordsClientException {
        if (reviewerInstructions != null) {
            HttpPost request = new HttpPost(baseUrlWithVersion + "/project/" + projectId + "/language/" + languageCode + "/reviewer-instructions");
            setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

            try {
                request.setEntity(new StringEntity(mapper.writeValueAsString(reviewerInstructions),UTF_8));
            } catch (Exception ex) {
                throw new CloudwordsClientException(ex);
            }
            return executeRequest(request, ReviewerInstructions.class);
        }
        return null;
    }

    @Override
    public ReviewerInstructions updateReviewerInstructions(ReviewerInstructions reviewerInstructions)
            throws CloudwordsClientException {
        if (reviewerInstructions != null) {
            HttpPut request = new HttpPut(baseUrlWithVersion + "/reviewer-instructions/" + reviewerInstructions.getId());
            setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

            try {
                request.setEntity(new StringEntity(mapper.writeValueAsString(reviewerInstructions),UTF_8));
            } catch (Exception ex) {
                throw new CloudwordsClientException(ex);
            }
            return executeRequest(request, ReviewerInstructions.class);
        }
        return null;
    }

    @Override
    public void deleteReviewerInstructions(int riId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpDelete(baseUrlWithVersion + "/reviewer-instructions/" + riId);
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        executeRequest(request);        
    }

    @Override
    public CustomerChangeOrder createChangeOrder(int projectId, CustomerChangeOrder customerChangeOrder)
            throws CloudwordsClientException {
        HttpPost request = new HttpPost(baseUrlWithVersion + "/project/" + projectId + "/changeOrder.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        try {
            request.setEntity(new StringEntity(mapper.writeValueAsString(customerChangeOrder), UTF_8));
        } catch (Exception ex) {
            throw new CloudwordsClientException(ex);
        }
        return executeRequest(request, CustomerChangeOrder.class);
    }

    @Override
    public CustomerChangeOrder getChangeOrder(int projectId, int changeOrderId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/changeOrder/"
                + changeOrderId + ".json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        TypeReference<CustomerChangeOrder> typeRef = new TypeReference<CustomerChangeOrder>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public List<CustomerChangeOrder> getChangeOrders(int projectId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/changeOrder.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        TypeReference<List<CustomerChangeOrder>> typeRef = new TypeReference<List<CustomerChangeOrder>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public CustomerChangeOrder cancelChangeOrder(CustomerChangeOrder customerChangeOrder)
            throws CloudwordsClientException {
        if (customerChangeOrder != null) {
            HttpPut request = new HttpPut(baseUrlWithVersion + "/project/" + customerChangeOrder.getProject().getId()
                    + "/changeOrder/" + customerChangeOrder.getId() + ".json");
            setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
            try {
                CodeAndDisplay status = new CodeAndDisplay();
                status.setCode("cancelled");
                customerChangeOrder.setStatus(status);
                request.setEntity(new StringEntity(mapper.writeValueAsString(customerChangeOrder), UTF_8));
            } catch (Exception ex) {
                throw new CloudwordsClientException(ex);
            }
            return executeRequest(request, CustomerChangeOrder.class);
        }
        return null;
    }
    
    @Override
    public CustomerChangeOrder submitChangeOrder(CustomerChangeOrder customerChangeOrder)
            throws CloudwordsClientException {
        if (customerChangeOrder != null) {
            HttpPut request = new HttpPut(baseUrlWithVersion + "/project/" + customerChangeOrder.getProject().getId()
                    + "/changeOrder/" + customerChangeOrder.getId() + ".json");
            setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
            try {
                CodeAndDisplay status = new CodeAndDisplay();
                status.setCode("requested");
                customerChangeOrder.setStatus(status);
                request.setEntity(new StringEntity(mapper.writeValueAsString(customerChangeOrder), UTF_8));
            } catch (Exception ex) {
                throw new CloudwordsClientException(ex);
            }
            return executeRequest(request, CustomerChangeOrder.class);
        }
        return null;
    }
    
    @Override
    public CustomerChangeOrder acceptChangeOrder(CustomerChangeOrder customerChangeOrder)
            throws CloudwordsClientException {
        if (customerChangeOrder != null) {
            HttpPut request = new HttpPut(baseUrlWithVersion + "/project/" + customerChangeOrder.getProject().getId()
                    + "/changeOrder/" + customerChangeOrder.getId() + ".json");
            setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
            try {
                CodeAndDisplay status = new CodeAndDisplay();
                status.setCode("customer_accepted");
                customerChangeOrder.setStatus(status);
                request.setEntity(new StringEntity(mapper.writeValueAsString(customerChangeOrder), UTF_8));
            } catch (Exception ex) {
                throw new CloudwordsClientException(ex);
            }
            return executeRequest(request, CustomerChangeOrder.class);
        }
        return null;
    }

    @Override
    public CloudwordsFile getChangeOrderAttachment(int projectId, int changeOrderId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/changeOrder/"
                + changeOrderId + "/file/attachment.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        return executeRequest(request, CloudwordsFile.class);
    }

    @Override
    public InputStream downloadChangeOrderFile(int projectId, int changeOrderId) throws CloudwordsClientException {
        InputStream stream = null;
        try {
            CloudwordsFile metadata = getChangeOrderAttachment(projectId, changeOrderId);
            stream = downloadFileFromMetadata(metadata);
            return stream;
        } catch (Exception ex) {
            try {
                if (stream != null) {
                    stream.close();
                }
            } catch (IOException ioe) {
                throw new CloudwordsClientException(ioe);
            }
            throw new CloudwordsClientException(ex);
        }
    }

    @Override
    public CloudwordsFile uploadChangeOrderFile(int projectId, int changeOrderId, File file)
            throws CloudwordsClientException {
        if (file != null) {
            return putFile("/project/" + projectId + "/changeOrder/"
                    + changeOrderId + "/file/attachment", file, CloudwordsFile.class);
        } else {
            throw new CloudwordsClientException("Unable to upload because no file was specified");
        }
    }

    @Override
    public CustomerChangeOrderBid getChangeOrderBid(int projectId, int changeOrderId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/changeOrder/" + changeOrderId + "/bid.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        return executeRequest(request, CustomerChangeOrderBid.class);
    }

    @Override
    public List<ProjectWorkflow> getProjectWorkflows(int projectId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/workflow.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        TypeReference<List<ProjectWorkflow>> typeRef = new TypeReference<List<ProjectWorkflow>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public ProjectWorkflow getProjectWorkflow(int projectId, String languageCode) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/workflow/" + languageCode + ".json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        return executeRequest(request, ProjectWorkflow.class);
    }

    @Override
    public ProjectWorkflow updateProjectWorkflow(int projectId, ProjectWorkflow workflow)
            throws CloudwordsClientException {
        HttpPut request = new HttpPut(baseUrlWithVersion + "/project/" + projectId
                + "/workflow/" + workflow.getLanguage().getLanguageCode() + ".json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        try {
            request.setEntity(new StringEntity(mapper.writeValueAsString(workflow), UTF_8));
        } catch (Exception ex) {
            throw new CloudwordsClientException(ex);
        }
        return executeRequest(request, ProjectWorkflow.class);
    }

    @Override
    public List<ProjectWorkflow> selectProjectWorkflow(int projectId, int templateId) throws CloudwordsClientException {
        Map<String,Integer> entity = new HashMap<String,Integer>();
        entity.put("workflowTemplateId", templateId);
        
        HttpPost request = new HttpPost(baseUrlWithVersion + "/project/" + projectId + "/workflow.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        try {
            request.setEntity(new StringEntity(mapper.writeValueAsString(entity), UTF_8));
        } catch (Exception ex) {
            throw new CloudwordsClientException(ex);
        }
        
        TypeReference<List<ProjectWorkflow>> typeRef = new TypeReference<List<ProjectWorkflow>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public List<WorkflowTemplate> getAvailableWorkflowTemplates(int projectId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/workflow-template.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        TypeReference<List<WorkflowTemplate>> typeRef = new TypeReference<List<WorkflowTemplate>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public WorkflowTemplate getWorkflowTemplate(int projectId, int templateId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/workflow-template/" + templateId + ".json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        return executeRequest(request, WorkflowTemplate.class);
    }

    @Override
    public List<ChangeOrderProjectWorkflow> getChangeOrderProjectWorkflows(int projectId, int changeOrderId)
            throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/changeOrder/" + changeOrderId + "/workflow.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        TypeReference<List<ChangeOrderProjectWorkflow>> typeRef = new TypeReference<List<ChangeOrderProjectWorkflow>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public ChangeOrderProjectWorkflow getChangeOrderProjectWorkflow(int projectId, int changeOrderId, String languageCode)
            throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/changeOrder/" + changeOrderId + "/workflow/" + languageCode + ".json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        return executeRequest(request, ChangeOrderProjectWorkflow.class);
    }

    @Override
    public ChangeOrderProjectWorkflow updateChangeOrderProjectWorkflow(int projectId, int changeOrderId,
            ChangeOrderProjectWorkflow workflow) throws CloudwordsClientException {
        HttpPut request = new HttpPut(baseUrlWithVersion + "/project/" + projectId + "/changeOrder/" + changeOrderId
                + "/workflow/" + workflow.getLanguage().getLanguageCode() + ".json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        try {
            request.setEntity(new StringEntity(mapper.writeValueAsString(workflow), UTF_8));
        } catch (Exception ex) {
            throw new CloudwordsClientException(ex);
        }
        return executeRequest(request, ChangeOrderProjectWorkflow.class);
    }
    
    @Override
    public SourceDocument addSourceDocument(int projectId, File file) throws CloudwordsClientException {
        return postFile("/project/" + projectId + "/file/source/document", file, SourceDocument.class);
    }

    @Override
    public List<SourceDocument> getSourceDocuments(int projectId) throws CloudwordsClientException {
        return getList("/project/" + projectId + "/file/source/document.json", SourceDocument.class);
    }

    @Override
    public SourceDocument getSourceDocument(int projectId, int documentId) throws CloudwordsClientException {
        return get("/project/" + projectId + "/file/source/document/" + documentId + ".json", SourceDocument.class);
    }

    @Override
    public SourceDocument replaceSourceDocument(int projectId, int documentId, File file)
            throws CloudwordsClientException {
        return putFile("/project/" + projectId + "/file/source/document/" + documentId, file, SourceDocument.class);
    }

    @Override
    public void deleteSourceDocument(int projectId, int documentId) throws CloudwordsClientException {
        delete("/project/" + projectId + "/file/source/document/" + documentId);
    }

    @Override
    public InputStream downloadSourceDocumentFile(int projectId, int documentId) throws CloudwordsClientException, IllegalStateException, IOException {
        SourceDocument sourceDocument = getSourceDocument(projectId, documentId);
        return downloadFileFromMetadata(sourceDocument.getFile());
    }

    @Override
    public InputStream downloadSourceDocumentXLIFF(int projectId, int documentId) throws CloudwordsClientException, IllegalStateException, IOException {
        SourceDocument sourceDocument = getSourceDocument(projectId, documentId);
        return downloadFileFromMetadata(sourceDocument.getXliff());
    }

    @Override
    public List<TranslatedDocument> getTranslatedDocuments(int projectId, Language language)
            throws CloudwordsClientException {
        return getList("/project/" + projectId + "/file/translated/language/" + language.getLanguageCode() + "/document.json", TranslatedDocument.class);
    }

    @Override
    public TranslatedDocument getTranslatedDocument(int projectId, Language language, int documentId)
            throws CloudwordsClientException {
        return get("/project/" + projectId + "/file/translated/language/" + language.getLanguageCode() + "/document/" + documentId + ".json", TranslatedDocument.class);
    }

    @Override
    public TranslatedDocument updateTranslatedDocument(int projectId, Language language, TranslatedDocument document)
            throws CloudwordsClientException {
        return put("/project/" + projectId + "/file/translated/language/" + language.getLanguageCode() + "/document/" + document.getId() + ".json", document);
    }

    @Override
    public TranslatedDocument addTranslatedDocument(int projectId, Language language, Integer sourceId, File file)
            throws CloudwordsClientException {
        HttpPost request = new HttpPost(baseUrlWithVersion + "/project/" + projectId + "/file/translated/language/" + language.getLanguageCode() + "/document");

        MultipartEntityBuilder entityBuilder = MultipartEntityBuilder.create().addPart(CW_FILE_PARAM, new FileBody(file));
        
        if (sourceId != null)   {
            entityBuilder.addPart(CW_SOURCE_PARAM, new StringBody(sourceId.toString(), ContentType.TEXT_PLAIN));
        }
        
        HttpEntity entity = entityBuilder.build();
        
        request.setEntity(entity);

        setBasicHeaders(request, JSON_CONTENT_TYPE, entity.getContentType().getValue());
        return executeRequest(request, TranslatedDocument.class);
    }

    @Override
    public TranslatedDocument replaceTranslatedDocument(int projectId, Language language, int documentId, File file)
            throws CloudwordsClientException {
        return putFile("/project/" + projectId + "/file/translated/language/" + language.getLanguageCode() + "/document/" + documentId, file, TranslatedDocument.class);
    }

    @Override
    public InputStream downloadTranslatedDocumentFile(int projectId, Language language, int documentId)
            throws CloudwordsClientException, IllegalStateException, IOException {
        return downloadFileFromMetadata(getTranslatedDocument(projectId, language, documentId).getFile());
    }

    @Override
    public InputStream downloadTranslatedDocumentXLIFF(int projectId, Language language, int documentId)
            throws CloudwordsClientException, IllegalStateException, IOException {
        return downloadFileFromMetadata(getTranslatedDocument(projectId, language, documentId).getXliff());
    }
    
    @Override
    public List<TranslatedDocument> updateTranslatedDocuments(int projectId, Language language,
            List<TranslatedDocument> documents) throws CloudwordsClientException {
        return putList("/project/" + projectId + "/file/translated/language/" + language.getLanguageCode() + "/document.json", TranslatedDocument.class, documents);
    }

    @Override
    public Webhook createWebhook(String eventType, String uri, Project project) throws CloudwordsClientException {
        if (project != null) {
            return createWebhook(eventType, uri, "project", project.getId(), null, null);
        } else {
            return null;
        }
    }
    
    @Override
    public Webhook createWebhook(String eventType, String uri, Project project, Language language) throws CloudwordsClientException {
        if (language != null) {
            return createWebhook(eventType, uri, "project", project.getId(), "language", language.getLanguageCode());
        } else {
            return createWebhook(eventType, uri, project);
        }
    }

    @Override
    public Webhook createWebhook(String eventType, String uri, String entityType, Integer entityId, String contextType, String contextId) throws CloudwordsClientException {
        if (eventType != null && uri != null) {
            Webhook webhook = new Webhook();
            webhook.setEventType(new CodeAndDisplay(eventType));
            webhook.setUri(uri);
            if( entityType != null && entityId != null ) {
                webhook.setEntityType(new CodeAndDisplay(entityType));
                webhook.setEntityId(entityId);
            }
            if( contextType != null && contextId != null ) {
                webhook.setContextType(new CodeAndDisplay(contextType));
                webhook.setContextId(contextId);
            }
            HttpPost request = new HttpPost(baseUrlWithVersion + "/webhook");
            setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
            try {
                request.setEntity(new StringEntity(mapper.writeValueAsString(webhook),UTF_8));
            } catch (Exception ex) {
                throw new CloudwordsClientException(ex);
            }
            return executeRequest(request, Webhook.class);
        }
        return null;
    }

    @Override
    public Webhook getWebhook(Integer webhookId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/webhook/" + webhookId + ".json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        return executeRequest(request, Webhook.class);
    }

    @Override
    public void deleteWebhook(Integer webhookId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpDelete(baseUrlWithVersion + "/webhook/" + webhookId);
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        executeRequest(request);       
    }

    @Override
    public List<Webhook> getWebhooks() throws CloudwordsClientException {
        return getList("/webhook.json", Webhook.class);
    }

    @Override
    public void addTranslatedDocumentPreview(int projectId, Language language, Integer translatedDocumentId, File file) throws CloudwordsClientException {
        putFile("/project/" + projectId + "/file/translated/language/" + language.getLanguageCode() + "/document/" + translatedDocumentId + "/preview", file);
        
    }

    @Override
    public void addSourceDocumentPreview(int projectId, Integer sourceDocumentId, File file) throws CloudwordsClientException {
        putFile("/project/" + projectId + "/file/source/document/" + sourceDocumentId + "/preview", file);
        
    }
    
    @Override
    public List<User> getAvailableProjectOwners() throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project-owner.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        TypeReference<List<User>> typeRef = new TypeReference<List<User>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public List<User> getAvailableProjectOwners(Department department) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/department/" + department.getId() + "/project-owner.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        TypeReference<List<User>> typeRef = new TypeReference<List<User>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public CustomerTranslationRequest getCurrentTranslationRequestForProject(int projectId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/translation-request/current.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        return executeRequest(request, CustomerTranslationRequest.class);
    }
    
    @Override
    public CustomerTranslationRequest requestTranslationRequestForProject(int projectId, CustomerTranslationRequest transRequest)
            throws CloudwordsClientException {
        HttpPost request = new HttpPost(baseUrlWithVersion + "/project/" + projectId + "/translation-request");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        try {
            request.setEntity(new StringEntity(mapper.writeValueAsString(transRequest), UTF_8));
        } catch (Exception ex) {
            throw new CloudwordsClientException(ex);
        }
        return executeRequest(request, CustomerTranslationRequest.class);
    }

    @Override
    public CustomerTranslationRequest selectWinningTranslationRequest(int projectId, int bidId) throws CloudwordsClientException {
        HttpPut request = new HttpPut(baseUrlWithVersion + "/project/" + projectId + "/translation-request");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        try {
            CustomerTranslationRequest transRequest = new CustomerTranslationRequest();
            transRequest.setRequiresBidding(true);
            BidOptions options = new BidOptions();
            options.setWinningBidId(bidId);
            transRequest.setOptions(options);
            request.setEntity(new StringEntity(mapper.writeValueAsString(transRequest),UTF_8));
        } catch (Exception ex) {
            throw new CloudwordsClientException(ex);
        }
        return executeRequest(request, CustomerTranslationRequest.class);
    }

    @Override
    public PagedResponse<Project> getProjects(PagedProjectQuery query) throws CloudwordsClientException {
        return get("/project.json", new TypeReference<PagedResponse<Project>>() {}, query);
    }

    @Override
    public PagedResponse<Project> getProjects(Department department, PagedProjectQuery query) throws CloudwordsClientException {
        if (department == null) {
            return getProjects(query);
        }
        
        return get("/department/" + department.getId() + "/project.json", new TypeReference<PagedResponse<Project>>() {}, query);
    }
}
