/*
 * Copyright 2012, Cloudwords, Inc.
 *
 * Licensed under the API LICENSE AGREEMENT, Version 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.cloudwords.com/developers/license-1.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.cloudwords.api.client;

import java.io.File;
import java.io.InputStream;
import java.util.List;

import com.cloudwords.api.client.exception.CloudwordsClientException;
import com.cloudwords.api.client.resources.ChangeOrder;
import com.cloudwords.api.client.resources.ChangeOrderBid;
import com.cloudwords.api.client.resources.CloudwordsFile;
import com.cloudwords.api.client.resources.CloudwordsLanguageFile;
import com.cloudwords.api.client.resources.CodeAndDisplay;
import com.cloudwords.api.client.resources.Language;
import com.cloudwords.api.client.resources.Project;
import com.cloudwords.api.client.resources.Task;
import com.cloudwords.api.client.resources.Vendor;
import com.cloudwords.api.client.resources.VendorBid;
import com.cloudwords.api.client.resources.VendorTranslationRequest;

/**
 * Client interface for the Cloudwords API. For more details and best practices please refer to the
 * Cloudwords API documentation found at <a
 * href="http://www.cloudwords.com/developers">Cloudwords</a>.
 * <p>
 * This API is designed for translation vendors on the Cloudwords marketplace that are looking to: 
 * <ul>
 * <li>Get a {@link Project}'s information</li>
 * <li>Download a {@link Project}'s source content and reference materials</li>
 * <li>Upload translated content for each of the {@link Project} language deliverable</li>
 * <li>Upload a translation memory for the translated {@link Project}</li>
 * </ul>
 * 
 * @author jonathansu
 * @since 1.3
 */
public interface CloudwordsVendorAPI extends BaseCloudwordsAPI {

    /**
     * Get the {@link Project}'s customer TM data {@link CloudwordsFile} resource for the specified
     * Cloudwords . The metadata will contain a contentPath for downloading the actual binary file.
     * 
     * @param projectId
     *            ID of the {@link Project} to returnTM data for
     * @return {@link Project} source {@link CloudwordsFile} resource
     * @throws CloudwordsClientException
     */
    public CloudwordsFile getProjectCustomerTM(int projectId) throws CloudwordsClientException;

    /**
     * Download the customer TM binary for the specified {@link Project} as an
     * {@link InputStream}.
     * <p>
     * <b>NOTE:</b> clients utilizing this method must handle the closing of {@link InputStream}
     * after use.
     * </p>
     * 
     * @param projectId
     *            ID of {@link Project} to download TM for
     * @return file {@link InputStream} if source file exists, null otherwise
     * @throws CloudwordsClientException
     */
    public InputStream downloadCustomerTMFile(int projectId) throws CloudwordsClientException;

    /**
     * Upload the translated file for a given Cloudwords {@link Project} {@link Language} deliverable. The provided
     * file must be a zip file.
     * <p>
     * Please note that this {@link Project} must already exists in Cloudwords, and the
     * {@link Project} must be in a state where translated file can be uploaded for the specified {@link Language}. 
     * Please refer to the API documentation for more information.
     * </p>
     * 
     * @param projectId
     *            ID of the {@link Project} to upload reference materials for
     * @param zipFile
     *            A single zip file to be uploaded for this {@link Project}
     * @param language
     *            The {@link Project} {@link Language} deliverable to upload this file for
     * @return Uploaded translated file {@link CloudwordsFile} resource
     * @throws CloudwordsClientException
     */
    public CloudwordsLanguageFile uploadProjectTranslatedFile(int projectId, Language language, File zipFile) throws CloudwordsClientException;

    /**
     * Upload the translated memory for a given Cloudwords {@link Project}. The provided
     * file must be a zip file and it must contain a valid TMX file for the given {@link Project}.
     * <p>
     * Please note that this {@link Project} must already exists in Cloudwords, and the
     * {@link Project} must be in a state where translated memory can be uploaded. 
     * Please refer to the API documentation for more information.
     * </p>
     * 
     * @param projectId
     *            ID of the {@link Project} to upload reference materials for
     * @param zipFile
     *            A single zip file to be uploaded for this {@link Project}
     * @return Uploaded translated memory {@link CloudwordsFile} resource
     * @throws CloudwordsClientException
     */
    public CloudwordsFile uploadProjectTranslationMemory(int projectId, File zipFile) throws CloudwordsClientException;
    
    /**
     * Get the specified {@link Project} {@link Task}.
     * 
     * @param projectId ID of the {@link Project} to get {@link Task} for
     * @param taskId ID of the {@link Task} to return
     * @return {@link Task} resource
     * @throws CloudwordsClientException
     */
    public Task getTask(int projectId,int taskId) throws CloudwordsClientException;
    
    /**
     * Get all {@link Task}s available to the user.
     * 
     * @return List of {@link Task}s
     * @throws CloudwordsClientException
     */
    public List<Task> getTasks() throws CloudwordsClientException;

    /**
     * Get all {@link Task}s by status.
     * 
     * @param taskStatus The {@link Task} status to filter by
     * @return List of {@link Task}s
     * @throws CloudwordsClientException
     */
    public List<Task> getTaskByStatus(CodeAndDisplay taskStatus) throws CloudwordsClientException;

    /**
     * Get all {@link Task}s for a project.
     * 
     * @param projectId ID of the {@link Project} to get {@link Task}s for
     * @return List of {@link Task}s
     * @throws CloudwordsClientException
     */
    public List<Task> getProjectTask(int projectId) throws CloudwordsClientException;
    
    /**
     * Get all {@link Task}s for a project by status.
     * 
     * @param projectId ID of the {@link Project} to get {@link Task}s for
     * @param taskStatus The {@link Task} status to filter by
     * @return List of {@link Task}s
     * @throws CloudwordsClientException
     */
    public List<Task> getProjectTaskByStatus(int projectId, CodeAndDisplay taskStatus) throws CloudwordsClientException;
    
    /**
     * Update a {@link Task} for a {@link Project}. 
     * 
     * @param task {@link Task} to update
     * @return The updated {@link Task}
     * @throws CloudwordsClientException
     */
    public Task updateTask(Task task) throws CloudwordsClientException;
    
    /**
     * Get the {@link Vendor} provided {@link VendorBid} for the specified Cloudwords {@link Project}.
     * 
     * @return the {@link Vendor} provided {@link VendorBid} for the specified Cloudwords {@link Project}.
     * @throws CloudwordsClientException
     */
    public VendorBid getBid(int projectId) throws CloudwordsClientException;
    
    /**
     * Accept {@link VendorBid} opportunity for the specified Cloudwords {@link Project}.
     * @param projectId ID of the {@link Project} to accept the bid for
     * @return The accepted {@link VendorBid}
     * @throws CloudwordsClientException
     */
    public VendorBid acceptBid(int projectId) throws CloudwordsClientException;
    
    /**
     * Reject {@link VendorBid} opportunity for the specified Cloudwords {@link Project}.
     * @param projectId ID of the {@link Project} to reject the bid for
     * @throws CloudwordsClientException
     */
    public void rejectBid(int projectId) throws CloudwordsClientException;   
    
    /**
     * Submit a {@link VendorBid} for the specified Cloudwords {@link Project}.
     * @param projectId ID of the {@link Project} to submit the bid for
     * @param bid The bid to submit
     * @return The submitted {@link VendorBid}
     * @throws CloudwordsClientException
     */
    public VendorBid submitBid(int projectId, VendorBid bid) throws CloudwordsClientException;

    /**
     * Gets the {@link ChangeOrder} associated with the unique {@link ChangeOrder} ID. Please refer
     * to the API documentation for more information.
     * @param projectId
     *            ID of the {@link Project} of the associated {@link ChangeOrder}
     * @param changeOrderId
     *            ID of the {@link ChangeOrder} to return
     * @return {@link ChangeOrder} resource
     * @throws CloudwordsClientException
     */
    public ChangeOrder getChangeOrder(int projectId, int changeOrderId) throws CloudwordsClientException;

    /**
     * Gets all {@link ChangeOrder}s for a project. Please refer to the API documentation for more information.
     * 
     * @param projectId
     *            ID of the {@link Project} to get {@link ChangeOrder}s for
     * @return List of {@link ChangeOrder}s
     * @throws CloudwordsClientException
     */
    public List<ChangeOrder> getChangeOrders(int projectId) throws CloudwordsClientException;

    /**
     * Updates the specified {@link ChangeOrder} by accepting the {@link ChangeOrder}. Accepting the {@link ChangeOrder}
     * doesn't submit any {@link ChangeOrderBid} to the customer. Please refer to the API documentation for more
     * information.
     * 
     * @param changeOrder
     *            {@link ChangeOrder} to update
     * @return {@link ChangeOrder} updated resource
     * @throws CloudwordsClientException
     */
    public ChangeOrder acceptChangeOrder(ChangeOrder changeOrder) throws CloudwordsClientException;
    
    /**
     * Updates the specified {@link ChangeOrder} by rejecting the {@link ChangeOrder}. Once you reject a
     * {@link ChangeOrder}, you can no longer view/submit any existing/new {@link ChangeOrderBid} to the customer.
     * Please refer to the API documentation for more information.
     * 
     * @param changeOrder
     *            {@link ChangeOrder} to update
     * @return {@link ChangeOrder} updated resource
     * @throws CloudwordsClientException
     */
    public ChangeOrder rejectChangeOrder(ChangeOrder changeOrder) throws CloudwordsClientException;
    
    /**
     * Updates the specified {@link ChangeOrder} by submitting the {@link ChangeOrder}. This action shall cause the
     * {@link ChangeOrderBid} associated with the {@link ChangeOrder} to be submitted to the customer. Please refer to
     * the API documentation for more information.
     * 
     * @param changeOrder
     *            {@link ChangeOrder} to update
     * @return {@link ChangeOrder} updated resource
     * @throws CloudwordsClientException
     */
    public ChangeOrder submitChangeOrder(ChangeOrder changeOrder) throws CloudwordsClientException;

    /**
     * Gets the specified {@link ChangeOrderBid} for the specified Cloudwords {@link ChangeOrder}. Please refer to the
     * API documentation for more information.
     * @param projectId
     *            ID of the {@link Project} of the associated {@link ChangeOrder}
     * @param changeOrderId
     *            ID of {@link ChangeOrder} to get the bid for
     * @return the specified {@link ChangeOrderBid} for the specified Cloudwords {@link ChangeOrder}.
     * @throws CloudwordsClientException
     */
    public ChangeOrderBid getChangeOrderBid(int projectId, int changeOrderId) throws CloudwordsClientException;

    /**
     * Updates the specified {@link ChangeOrderBid} for the {@link ChangeOrder} before submitting. The bid won't be
     * submitted until you submit the {@link ChangeOrder}. You can update the update the {@link ChangeOrderBid} as many
     * times as you want before doing the final {@link ChangeOrder}. Please refer to the API documentation for more
     * information.
     * @param projectId
     *            ID of the {@link Project} of the associated {@link ChangeOrder}
     * @param changeOrderId
     *            ID of {@link ChangeOrder} of the associated {@link ChangeOrderBid}
     * @param changeOrderBid
     *            {@link ChangeOrderBid} to to be updated
     * @return the specified {@link ChangeOrderBid} for the specified Cloudwords {@link ChangeOrder}.
     * @throws CloudwordsClientException
     */
    public ChangeOrderBid updateChangeOrderBid(int projectId, int changeOrderId, ChangeOrderBid changeOrderBid) throws CloudwordsClientException;

    /**
     * Get the current {@link VendorTranslationRequest} for the specified {@link Project}.
     * 
     * @param projectId
     *            ID of the {@link Project} to get the {@link VendorTranslationRequest} for
     * @return current {@link VendorTranslationRequest} resource for the {@link Project} if requests have already been
     *         requested for this {@link Project}
     * @throws CloudwordsClientException
     */
    public VendorTranslationRequest getCurrentTranslationRequestForProject(int projectId) throws CloudwordsClientException;

    /**
     * Updates the specified {@link VendorTranslationRequest} by accepting the {@link VendorTranslationRequest}.
     * Please refer to the API documentation for more information.
     * 
     * @param projectId
     *            ID of the {@link Project} to request translation for
     * @param transRequest
     *            A {@link VendorTranslationRequest}s to accept
     * @throws CloudwordsClientException
     */
    public VendorTranslationRequest acceptTranslationRequest(int projectId, VendorTranslationRequest transRequest) 
            throws CloudwordsClientException;
    
    /**
     * Updates the specified {@link VendorTranslationRequest} by rejecting the {@link VendorTranslationRequest}.
     * Please refer to the API documentation for more information.
     * 
     * @param projectId
     *            ID of the {@link Project} to request translation for
     * @param transRequest
     *            A {@link VendorTranslationRequest}s to reject
     * @throws CloudwordsClientException
     */
    public VendorTranslationRequest rejectTranslationRequest(int projectId, VendorTranslationRequest transRequest) 
            throws CloudwordsClientException;
    
    /**
     * Submit a {@link VendorTranslationRequest} for the specified Cloudwords {@link Project}.
     * @param projectId ID of the {@link Project} to submit the bid for
     * @param transRequest A {@link VendorTranslationRequest}s containing the bid to submit
     * @return The submitted {@link VendorTranslationRequest}
     * @throws CloudwordsClientException
     */
    public VendorTranslationRequest submitTranslationRequest(int projectId, VendorTranslationRequest transRequest) throws CloudwordsClientException;

}
