/*
 * Copyright 2012, Cloudwords, Inc.
 *
 * Licensed under the API LICENSE AGREEMENT, Version 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.cloudwords.com/developers/license-1.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.cloudwords.api.client;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.List;

import org.apache.commons.lang3.StringUtils;
import org.apache.http.HttpEntity;
import org.apache.http.HttpHost;
import org.apache.http.client.CredentialsProvider;
import org.apache.http.client.HttpClient;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.client.methods.HttpDelete;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.methods.HttpPut;
import org.apache.http.client.methods.HttpUriRequest;
import org.apache.http.config.SocketConfig;
import org.apache.http.entity.ContentType;
import org.apache.http.entity.StringEntity;
import org.apache.http.entity.mime.MultipartEntity;
import org.apache.http.entity.mime.MultipartEntityBuilder;
import org.apache.http.entity.mime.content.FileBody;
import org.apache.http.entity.mime.content.StringBody;
import org.apache.http.impl.client.HttpClientBuilder;

import com.cloudwords.api.client.exception.CloudwordsClientException;
import com.cloudwords.api.client.resources.ChangeOrder;
import com.cloudwords.api.client.resources.ChangeOrderBid;
import com.cloudwords.api.client.resources.ChangeOrderProjectWorkflow;
import com.cloudwords.api.client.resources.CloudwordsFile;
import com.cloudwords.api.client.resources.CloudwordsLanguageFile;
import com.cloudwords.api.client.resources.CodeAndDisplay;
import com.cloudwords.api.client.resources.Language;
import com.cloudwords.api.client.resources.PagedProjectQuery;
import com.cloudwords.api.client.resources.PagedResponse;
import com.cloudwords.api.client.resources.Project;
import com.cloudwords.api.client.resources.ProjectWorkflow;
import com.cloudwords.api.client.resources.SourceDocument;
import com.cloudwords.api.client.resources.Task;
import com.cloudwords.api.client.resources.TranslatedDocument;
import com.cloudwords.api.client.resources.VendorBid;
import com.cloudwords.api.client.resources.VendorTranslationRequest;
import com.cloudwords.api.client.resources.Webhook;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;

/**
 * Basic implementation of the Cloudwords Vendor API client.
 * 
 * @author jonathansu
 * @since 1.3
 */
public class CloudwordsVendorClient extends BaseCloudwordsClient implements CloudwordsVendorAPI {
    
    // The authorization token to validate identify when accessing the Cloudwords API
    private final String authToken;
    // The domain + version for connecting to the API (e.g. https://api.cloudwords.com/v1)
    private final ObjectMapper mapper = new ObjectMapper();
    // The main HTTP client that will send/receive communication with the Cloudwords API
    private HttpClient client;

    /**
     * Convenience constructor that provides default configuration.
     * 
     * @param baseAPIURL
     *            The base domain of the Cloudwords API
     * @param version
     *            The version of the Cloudwords API to use
     * @param authorizationToken
     *            The authorization token to validate identify when accessing the Cloudwords API
     */
    public CloudwordsVendorClient(String baseAPIURL, String version, String authorizationToken) {
        this(baseAPIURL, version, authorizationToken, connectionTimeoutMillis, socketTimeoutMillis, maxTotalConnections);
    }

    
    /**
     * Convenience constructor that provides default configuration.
     * 
     * @param baseAPIURL
     *            The base domain of the Cloudwords API
     * @param version
     *            The version of the Cloudwords API to use
     * @param authorizationToken
     *            The authorization token to validate identify when accessing the Cloudwords API
     * @param proxy
     *            The proxy server to use for connections
     * @param proxyCredentialsProvider
     *            The credentials provider to use with the proxy server
     */
    public CloudwordsVendorClient(String baseAPIURL, String version, String authorizationToken, HttpHost proxy, CredentialsProvider proxyCredentialsProvider) {
        this(baseAPIURL, version, authorizationToken, connectionTimeoutMillis, socketTimeoutMillis, maxTotalConnections, proxy, proxyCredentialsProvider);
    }
    
    /**
     * Base constructor.
     * 
     * @param baseAPIURL
     *            The base domain of the Cloudwords API
     * @param version
     *            The version of the Cloudwords API to use
     * @param authorizationToken
     *            The authorization token to validate identify when accessing the Cloudwords API
     * @param connectionTimeoutMillis
     *            Determines the timeout in milliseconds until a connection is established. A
     *            timeout value of zero is interpreted as an infinite timeout.
     * @param socketTimeoutMillis
     *            Defines the socket timeout (<code>SO_TIMEOUT</code>) in milliseconds, which is the
     *            timeout for waiting for data or, put differently, a maximum period inactivity
     *            between two consecutive data packets). A timeout value of zero is interpreted as
     *            an infinite timeout.
     * @param maxTotalConnections
     *            The max concurrent connections the client can establish against the Cloudwords API
     */
    public CloudwordsVendorClient(String baseAPIURL, String version, String authorizationToken, int connectionTimeoutMillis,
            int socketTimeoutMillis, int maxTotalConnections, HttpHost proxy, CredentialsProvider proxyCredentialsProvider) {
        
        this(baseAPIURL, version, authorizationToken, null);
        
        if (maxTotalConnections > 5) {
            throw new IllegalArgumentException("Only 5 concurrent connections per Cloudwords account is supported");
        }

        HttpClientBuilder clientBuilder = HttpClientBuilder.create()
            .setDefaultSocketConfig(SocketConfig.custom().setSoTimeout(socketTimeoutMillis).build())
            .setDefaultRequestConfig(RequestConfig.custom().setConnectTimeout(connectionTimeoutMillis).setSocketTimeout(socketTimeoutMillis).build())
            .setMaxConnTotal(maxTotalConnections);
        
        if (proxy != null) {
            clientBuilder.setProxy(proxy);
            if (proxyCredentialsProvider != null) {
                clientBuilder.setDefaultCredentialsProvider(proxyCredentialsProvider);
            }
        }
        
        this.client = clientBuilder.build();
    }
    
    /**
     * Creates a new client that takes a user-provided instance of HttpClient. The caller is responsible for ensuring this client is properly initialized.
     * As of HttpClient 4.3.5, the following code is used by the other constructors in this class to build a proper client:
     * <p>
     * HttpClientBuilder.create()
            .setDefaultSocketConfig(SocketConfig.custom().setSoTimeout(socketTimeoutMillis).build())
            .setDefaultRequestConfig(RequestConfig.custom().setConnectTimeout(connectionTimeoutMillis).setSocketTimeout(socketTimeoutMillis).build())
            .setMaxConnTotal(maxTotalConnections).build();
        </p>
     * 
     * @param baseAPIURL The base domain of the Cloudwords API
     * @param version The version of the Cloudwords API to use
     * @param authorizationToken The authorization token to validate identify when accessing the Cloudwords API
     * @param client The HTTP Client instance to use
     */
    public CloudwordsVendorClient(String baseAPIURL, String version, String authorizationToken, HttpClient client)    {
        this.apiVersion = Double.parseDouble(version);
        this.authToken = new StringBuilder("UserToken ").append(authorizationToken).toString();
        this.baseUrlWithVersion = baseAPIURL + "/" + version;
        this.client = client;
        initObjectMapper();
    }
    
    /**
     * Base constructor.
     * 
     * @param baseAPIURL
     *            The base domain of the Cloudwords API
     * @param version
     *            The version of the Cloudwords API to use
     * @param authorizationToken
     *            The authorization token to validate identify when accessing the Cloudwords API
     * @param connectionTimeoutMillis
     *            Determines the timeout in milliseconds until a connection is established. A
     *            timeout value of zero is interpreted as an infinite timeout.
     * @param socketTimeoutMillis
     *            Defines the socket timeout (<code>SO_TIMEOUT</code>) in milliseconds, which is the
     *            timeout for waiting for data or, put differently, a maximum period inactivity
     *            between two consecutive data packets). A timeout value of zero is interpreted as
     *            an infinite timeout.
     * @param maxTotalConnections
     *            The max concurrent connections the client can establish against the Cloudwords API
     */
    public CloudwordsVendorClient(String baseAPIURL, String version, String authorizationToken, int connectionTimeoutMillis,
            int socketTimeoutMillis, int maxTotalConnections) {
        
        this(baseAPIURL, version, authorizationToken, connectionTimeoutMillis, socketTimeoutMillis, maxTotalConnections, null, null);
    }

    @Override
    public Project getProject(int projectId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + ".json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        return executeRequest(request, Project.class);
    }

    @Override
    public CloudwordsFile getProjectSource(int projectId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/file/source.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        return executeRequest(request, CloudwordsFile.class);
    }

    @Override
    public InputStream downloadSourceFile(int projectId) throws CloudwordsClientException {
        InputStream stream = null;
        try {
            CloudwordsFile metadata = getProjectSource(projectId);
            stream = downloadFileFromMetadata(metadata);
            return stream;
        } catch (Exception ex) {
            try {
                if(stream != null) {
                    stream.close();                 
                }
            } catch (IOException ioe) {
                throw new CloudwordsClientException(ioe);
            }
            throw new CloudwordsClientException(ex);
        }
    }
    
    @Override
    public CloudwordsFile getProjectCustomerTM(int projectId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/file/customer-tm.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        return executeRequest(request, CloudwordsFile.class);
    }

    @Override
    public InputStream downloadCustomerTMFile(int projectId) throws CloudwordsClientException {
        InputStream stream = null;
        try {
            CloudwordsFile metadata = getProjectCustomerTM(projectId);
            stream = downloadFileFromMetadata(metadata);
            return stream;
        } catch (Exception ex) {
            try {
                if(stream != null) {
                    stream.close();                 
                }
            } catch (IOException ioe) {
                throw new CloudwordsClientException(ioe);
            }
            throw new CloudwordsClientException(ex);
        }
    }

    @Override
    public List<CloudwordsFile> getProjectReferences(int projectId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/file/reference.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        TypeReference<List<CloudwordsFile>> typeRef = new TypeReference<List<CloudwordsFile>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public CloudwordsFile getProjectReference(int projectId, int documentId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/file/reference/"
                + documentId + ".json");

        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        return executeRequest(request, CloudwordsFile.class);
    }

    @Override
    public InputStream downloadReferenceFile(int projectId, int documentId) throws CloudwordsClientException {
        InputStream stream = null;
        try {
            CloudwordsFile metadata = getProjectReference(projectId, documentId);
            stream = downloadFileFromMetadata(metadata);
            return stream;
        } catch (Exception ex) {
            try {
                if(stream != null) {
                    stream.close();                 
                }
            } catch (IOException ioe) {
                throw new CloudwordsClientException(ioe);
            }
            throw new CloudwordsClientException(ex);
        }
    }

    @Override
    public CloudwordsLanguageFile uploadProjectTranslatedFile(int projectId, Language language, File zipFile) throws CloudwordsClientException {
        if (zipFile != null) {
            validateZipFile(zipFile);
            
            if (language != null && language.getLanguageCode() != null) {
                HttpPut request = new HttpPut(baseUrlWithVersion + "/project/" + projectId
                        + "/file/translated/language/" + language.getLanguageCode());
                HttpEntity entity = MultipartEntityBuilder.create().addPart(CW_FILE_PARAM, new FileBody(zipFile)).build();
                request.setEntity(entity);
                
                setBasicHeaders(request, JSON_CONTENT_TYPE, entity.getContentType()
                        .getValue());
                
                CloudwordsLanguageFile output = executeRequest(request,
                        CloudwordsLanguageFile.class);

                return output;
            } else {
                throw new CloudwordsClientException(
                        "Unable to upload project translated file for language because either the language object or the language code is null");
            }
        } else {
            throw new CloudwordsClientException(
                    "Unable to upload project translated file because no file was specified. Project id: '"
                            + projectId + "'");
        }
    }
    
    @Override
    public CloudwordsFile getMasterProjectTranslatedFile(int projectId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/file/translated.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        return executeRequest(request, CloudwordsFile.class);
    }

    @Override
    public List<CloudwordsLanguageFile> getProjectTranslatedFiles(int projectId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId
                + "/file/translated/language.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        TypeReference<List<CloudwordsLanguageFile>> typeRef = new TypeReference<List<CloudwordsLanguageFile>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public CloudwordsLanguageFile getProjectTranslatedFile(int projectId, Language language)
            throws CloudwordsClientException {
        if (language != null && language.getLanguageCode() != null) {
            HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId
                    + "/file/translated/language/" + language.getLanguageCode() + ".json");

            setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
            return executeRequest(request, CloudwordsLanguageFile.class);
        } else {
            throw new CloudwordsClientException(
                    "Unable to get project translated file for language because either the language object or the language code is null");
        }
    }

    @Override
    public InputStream downloadMasterTranslatedFile(int projectId) throws CloudwordsClientException {
        InputStream stream = null;
        try {
            CloudwordsFile metadata = getMasterProjectTranslatedFile(projectId);
            stream = downloadFileFromMetadata(metadata);
            return stream;
        } catch (Exception ex) {
            try {
                if(stream != null) {
                    stream.close();                 
                }
            } catch (IOException ioe) {
                throw new CloudwordsClientException(ioe);
            }
            throw new CloudwordsClientException(ex);
        }
    }

    @Override
    public InputStream downloadTranslatedFile(int projectId, Language language) throws CloudwordsClientException {
        InputStream stream = null;
        try {
            CloudwordsLanguageFile metadata = getProjectTranslatedFile(projectId, language);
            stream = downloadFileFromMetadata(metadata);
            return stream;
        } catch (Exception ex) {
            try {
                if(stream != null) {
                    stream.close();                 
                }
            } catch (IOException ioe) {
                throw new CloudwordsClientException(ioe);
            }
            throw new CloudwordsClientException(ex);
        }
    }

    @Override
    public CloudwordsFile uploadProjectTranslationMemory(int projectId, File zipFile) throws CloudwordsClientException {
        if (zipFile != null) {
            validateZipFile(zipFile);
            
            HttpPut request = new HttpPut(baseUrlWithVersion + "/project/" + projectId + "/file/tm");

            MultipartEntity entity = new MultipartEntity();
            entity.addPart(CW_FILE_PARAM, new FileBody(zipFile));
            request.setEntity(entity);

            setBasicHeaders(request, JSON_CONTENT_TYPE, entity.getContentType()
                    .getValue());
            CloudwordsFile output = executeRequest(request,
                    CloudwordsFile.class);

            return output;
        } else {
            throw new CloudwordsClientException(
                    "Unable to upload project translation memory because no file was specified. Project id: '"
                            + projectId + "'");
        }
    }

    @Override
    public CloudwordsFile getProjectTranslationMemory(int projectId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/file/tm.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        return executeRequest(request, CloudwordsFile.class);
    }
    
    @Override
    public InputStream downloadProjectTranslationMemory(int projectId) throws CloudwordsClientException {
        InputStream stream = null;
        try {
            CloudwordsFile metadata = getProjectTranslationMemory(projectId);
            stream = downloadFileFromMetadata(metadata);
            return stream;
        } catch (Exception ex) {
            try {
                if(stream != null) {
                    stream.close();                 
                }
            } catch (IOException ioe) {
                throw new CloudwordsClientException(ioe);
            }
            throw new CloudwordsClientException(ex);
        }
    }

    @Override
    public List<Project> getOpenProjects() throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/open.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        TypeReference<List<Project>> typeRef = new TypeReference<List<Project>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public List<Project> getClosedProjects() throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/closed.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        TypeReference<List<Project>> typeRef = new TypeReference<List<Project>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    protected HttpClient getClient() {
        return this.client;
    }

    @Override
    protected ObjectMapper getMapper() {
        return this.mapper;
    }

    @Override
    protected String getAuthToken() {
        return this.authToken;
    }

    @Override
    public Task getTask(int projectId, int taskId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/task/" + taskId + ".json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        TypeReference<Task> typeRef = new TypeReference<Task>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public List<Task> getTasks() throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/task.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        TypeReference<List<Task>> typeRef = new TypeReference<List<Task>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public List<Task> getTaskByStatus(CodeAndDisplay taskStatus) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/task/status/" + taskStatus.getCode() + ".json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        TypeReference<List<Task>> typeRef = new TypeReference<List<Task>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public List<Task> getProjectTask(int projectId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/task.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        TypeReference<List<Task>> typeRef = new TypeReference<List<Task>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public List<Task> getProjectTaskByStatus(int projectId, CodeAndDisplay taskStatus) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/task/status/"
                + taskStatus.getCode() + ".json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        TypeReference<List<Task>> typeRef = new TypeReference<List<Task>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public Task updateTask(Task task) throws CloudwordsClientException {
        if (task != null) {
            HttpPut request = new HttpPut(baseUrlWithVersion + "/project/" + task.getProject().getId() + "/task/"
                    + task.getId() + ".json");
            setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
            try {
                request.setEntity(new StringEntity(mapper.writeValueAsString(task),CloudwordsCustomerClient.UTF_8));
            } catch (Exception ex) {
                throw new CloudwordsClientException(ex);
            }
            return executeRequest(request, Task.class);
        }
        return null;
    }
    
    @Override
    public Project updateProject(Project project) throws CloudwordsClientException {
        if (project != null) {
            HttpPut request = new HttpPut(baseUrlWithVersion + "/project/" + project.getId() + ".json");
            setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
            try {
                request.setEntity(new StringEntity(mapper.writeValueAsString(project),CloudwordsCustomerClient.UTF_8));
            } catch (Exception ex) {
                throw new CloudwordsClientException(ex);
            }
            return executeRequest(request, Project.class);
        }
        return null;
    }

    @Override
    public CloudwordsFile getTaskAttachment(int projectId,int taskId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/"+projectId+"/task/" + taskId
                + "/file/attachment.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        return executeRequest(request, CloudwordsFile.class);
    }

    @Override
    public InputStream downloadTaskAttachment(int projectId,int taskId) throws CloudwordsClientException {
        InputStream stream = null;
        try {
            CloudwordsFile metadata = getTaskAttachment(projectId, taskId);
            stream = downloadFileFromMetadata(metadata);
            return stream;
        } catch (Exception ex) {
            try {
                if (stream != null) {
                    stream.close();
                }
            } catch (IOException ioe) {
                throw new CloudwordsClientException(ioe);
            }
            throw new CloudwordsClientException(ex);
        }
    }

    @Override
    public VendorBid getBid(int projectId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/bid.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        return executeRequest(request, VendorBid.class);
    }

    @Override
    public VendorBid acceptBid(int projectId) throws CloudwordsClientException {
        HttpPost request = new HttpPost(baseUrlWithVersion + "/project/" + projectId + "/bid.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        try {
            VendorBid bid = new VendorBid();
            CodeAndDisplay status = new CodeAndDisplay();
            status.setCode("vendor_accepted");
            bid.setStatus(status);
            request.setEntity(new StringEntity(mapper.writeValueAsString(bid), CloudwordsCustomerClient.UTF_8));
        } catch (Exception ex) {
            throw new CloudwordsClientException(ex);
        }
        return executeRequest(request, VendorBid.class);
    }
    
    @Override
    public void rejectBid(int projectId) throws CloudwordsClientException {
        HttpPost request = new HttpPost(baseUrlWithVersion + "/project/" + projectId + "/bid.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        try {
            VendorBid bid = new VendorBid();
            CodeAndDisplay status = new CodeAndDisplay();
            status.setCode("vendor_rejected");
            bid.setStatus(status);
            request.setEntity(new StringEntity(mapper.writeValueAsString(bid), CloudwordsCustomerClient.UTF_8));
        } catch (Exception ex) {
            throw new CloudwordsClientException(ex);
        }
        executeRequest(request);
    }

    @Override
    public VendorBid submitBid(int projectId, VendorBid bid) throws CloudwordsClientException {
        HttpPut request = new HttpPut(baseUrlWithVersion + "/project/" + projectId + "/bid.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        try {
            CodeAndDisplay status = new CodeAndDisplay();
            status.setCode("submitted");
            bid.setStatus(status);
            request.setEntity(new StringEntity(mapper.writeValueAsString(bid), CloudwordsCustomerClient.UTF_8));
        } catch (Exception ex) {
            throw new CloudwordsClientException(ex);
        }
        return executeRequest(request, VendorBid.class);
    }

    @Override
    public ChangeOrder getChangeOrder(int projectId, int changeOrderId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/changeOrder/"
                + changeOrderId + ".json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        TypeReference<ChangeOrder> typeRef = new TypeReference<ChangeOrder>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public List<ChangeOrder> getChangeOrders(int projectId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/changeOrder.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        TypeReference<List<ChangeOrder>> typeRef = new TypeReference<List<ChangeOrder>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public CloudwordsFile getChangeOrderAttachment(int projectId, int changeOrderId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/changeOrder/"
                + changeOrderId + "/file/attachment.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        return executeRequest(request, CloudwordsFile.class);
    }

    @Override
    public InputStream downloadChangeOrderFile(int projectId, int changeOrderId) throws CloudwordsClientException {
        InputStream stream = null;
        try {
            CloudwordsFile metadata = getChangeOrderAttachment(projectId, changeOrderId);
            stream = downloadFileFromMetadata(metadata);
            return stream;
        } catch (Exception ex) {
            try {
                if (stream != null) {
                    stream.close();
                }
            } catch (IOException ioe) {
                throw new CloudwordsClientException(ioe);
            }
            throw new CloudwordsClientException(ex);
        }
    }

    @Override
    public ChangeOrderBid getChangeOrderBid(int projectId, int changeOrderId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/changeOrder/"
                + changeOrderId + "/bid.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        return executeRequest(request, ChangeOrderBid.class);
    }

    @Override
    public ChangeOrderBid updateChangeOrderBid(int projectId, int changeOrderId, ChangeOrderBid changeOrderBid)
            throws CloudwordsClientException {
        if (changeOrderBid != null) {
            HttpPut request = new HttpPut(baseUrlWithVersion + "/project/" + projectId + "/changeOrder/"
                    + changeOrderId + "/bid.json");
            setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
            try {
                request.setEntity(new StringEntity(mapper.writeValueAsString(changeOrderBid),
                        CloudwordsCustomerClient.UTF_8));
            } catch (Exception ex) {
                throw new CloudwordsClientException(ex);
            }
            return executeRequest(request, ChangeOrderBid.class);
        }
        return null;
    }

    @Override
    public ChangeOrder acceptChangeOrder(ChangeOrder changeOrder) throws CloudwordsClientException {
        if (changeOrder != null) {
            HttpPut request = new HttpPut(baseUrlWithVersion + "/project/" + changeOrder.getProject().getId()
                    + "/changeOrder/" + changeOrder.getId() + ".json");
            setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
            try {
                CodeAndDisplay status = new CodeAndDisplay();
                status.setCode("vendor_accepted");
                changeOrder.setStatus(status);
                request.setEntity(new StringEntity(mapper.writeValueAsString(changeOrder),
                        CloudwordsCustomerClient.UTF_8));
            } catch (Exception ex) {
                throw new CloudwordsClientException(ex);
            }
            return executeRequest(request, ChangeOrder.class);
        }
        return null;
    }

    @Override
    public ChangeOrder rejectChangeOrder(ChangeOrder changeOrder) throws CloudwordsClientException {
        if (changeOrder != null) {
            HttpPut request = new HttpPut(baseUrlWithVersion + "/project/" + changeOrder.getProject().getId()
                    + "/changeOrder/" + changeOrder.getId() + ".json");
            setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
            try {
                CodeAndDisplay status = new CodeAndDisplay();
                status.setCode("vendor_rejected");
                changeOrder.setStatus(status);
                request.setEntity(new StringEntity(mapper.writeValueAsString(changeOrder),
                        CloudwordsCustomerClient.UTF_8));
            } catch (Exception ex) {
                throw new CloudwordsClientException(ex);
            }
            return executeRequest(request, ChangeOrder.class);
        }
        return null;
    }

    @Override
    public ChangeOrder submitChangeOrder(ChangeOrder changeOrder) throws CloudwordsClientException {
        if (changeOrder != null) {
            HttpPut request = new HttpPut(baseUrlWithVersion + "/project/" + changeOrder.getProject().getId()
                    + "/changeOrder/" + changeOrder.getId() + ".json");
            setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
            try {
                CodeAndDisplay status = new CodeAndDisplay();
                status.setCode("vendor_bid");
                changeOrder.setStatus(status);
                request.setEntity(new StringEntity(mapper.writeValueAsString(changeOrder),
                        CloudwordsCustomerClient.UTF_8));
            } catch (Exception ex) {
                throw new CloudwordsClientException(ex);
            }
            return executeRequest(request, ChangeOrder.class);
        }
        return null;
    }
    
    @Override
    public List<ProjectWorkflow> getProjectWorkflows(int projectId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/workflow.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        TypeReference<List<ProjectWorkflow>> typeRef = new TypeReference<List<ProjectWorkflow>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public ProjectWorkflow getProjectWorkflow(int projectId, String languageCode) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/workflow/" + languageCode + ".json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        return executeRequest(request, ProjectWorkflow.class);
    }

    @Override
    public List<ChangeOrderProjectWorkflow> getChangeOrderProjectWorkflows(int projectId, int changeOrderId)
            throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/changeOrder/" + changeOrderId + "/workflow.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        TypeReference<List<ChangeOrderProjectWorkflow>> typeRef = new TypeReference<List<ChangeOrderProjectWorkflow>>() {};
        return executeRequest(request, typeRef);
    }

    @Override
    public ChangeOrderProjectWorkflow getChangeOrderProjectWorkflow(int projectId, int changeOrderId, String languageCode)
            throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/changeOrder/" + changeOrderId + "/workflow/" + languageCode + ".json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);

        return executeRequest(request, ChangeOrderProjectWorkflow.class);
    }

    @Override
    public List<SourceDocument> getSourceDocuments(int projectId) throws CloudwordsClientException {
        return getList("/project/" + projectId + "/file/source/document.json", SourceDocument.class);
    }

    @Override
    public SourceDocument getSourceDocument(int projectId, int documentId) throws CloudwordsClientException {
        return get("/project/" + projectId + "/file/source/document/" + documentId + ".json", SourceDocument.class);
    }

    @Override
    public InputStream downloadSourceDocumentFile(int projectId, int documentId) throws CloudwordsClientException, IllegalStateException, IOException {
        SourceDocument sourceDocument = getSourceDocument(projectId, documentId);
        return downloadFileFromMetadata(sourceDocument.getFile());
    }

    @Override
    public InputStream downloadSourceDocumentXLIFF(int projectId, int documentId) throws CloudwordsClientException, IllegalStateException, IOException {
        SourceDocument sourceDocument = getSourceDocument(projectId, documentId);
        return downloadFileFromMetadata(sourceDocument.getXliff());
    }

    @Override
    public List<TranslatedDocument> getTranslatedDocuments(int projectId, Language language)
            throws CloudwordsClientException {
        return getList("/project/" + projectId + "/file/translated/language/" + language.getLanguageCode() + "/document.json", TranslatedDocument.class);
    }

    @Override
    public TranslatedDocument getTranslatedDocument(int projectId, Language language, int documentId)
            throws CloudwordsClientException {
        return get("/project/" + projectId + "/file/translated/language/" + language.getLanguageCode() + "/document/" + documentId + ".json", TranslatedDocument.class);
    }

    @Override
    public TranslatedDocument updateTranslatedDocument(int projectId, Language language, TranslatedDocument document)
            throws CloudwordsClientException {
        return put("/project/" + projectId + "/file/translated/language/" + language.getLanguageCode() + "/document/" + document.getId() + ".json", document);
    }

    @Override
    public TranslatedDocument addTranslatedDocument(int projectId, Language language, Integer sourceId, File file)
            throws CloudwordsClientException {
        HttpPost request = new HttpPost(baseUrlWithVersion + "/project/" + projectId + "/file/translated/language/" + language.getLanguageCode() + "/document");

        MultipartEntityBuilder entityBuilder = MultipartEntityBuilder.create().addPart(CW_FILE_PARAM, new FileBody(file));
        
        if (sourceId != null)   {
            entityBuilder.addPart(CW_SOURCE_PARAM, new StringBody(sourceId.toString(), ContentType.TEXT_PLAIN));
        }
        
        HttpEntity entity = entityBuilder.build();
        
        request.setEntity(entity);

        setBasicHeaders(request, JSON_CONTENT_TYPE, entity.getContentType().getValue());
        return executeRequest(request, TranslatedDocument.class);
    }

    @Override
    public TranslatedDocument replaceTranslatedDocument(int projectId, Language language, int documentId, File file)
            throws CloudwordsClientException {
        return putFile("/project/" + projectId + "/file/translated/language/" + language.getLanguageCode() + "/document/" + documentId, file, TranslatedDocument.class);
    }

    @Override
    public InputStream downloadTranslatedDocumentFile(int projectId, Language language, int documentId)
            throws CloudwordsClientException, IllegalStateException, IOException {
        return downloadFileFromMetadata(getTranslatedDocument(projectId, language, documentId).getFile());
    }

    @Override
    public InputStream downloadTranslatedDocumentXLIFF(int projectId, Language language, int documentId)
            throws CloudwordsClientException, IllegalStateException, IOException {
        return downloadFileFromMetadata(getTranslatedDocument(projectId, language, documentId).getXliff());

    }

    @Override
    public CloudwordsLanguageFile approveTranslatedFile(int projectId, Language language)
            throws CloudwordsClientException {
        CloudwordsLanguageFile file = new CloudwordsLanguageFile();
        CodeAndDisplay status = new CodeAndDisplay();
        status.setCode("approved");
        file.setStatus(status);
        return put("/project/" + projectId + "/file/translated/language/" + language.getLanguageCode() + ".json", file);
    }
    
    @Override
    public List<TranslatedDocument> updateTranslatedDocuments(int projectId, Language language,
            List<TranslatedDocument> documents) throws CloudwordsClientException {
        return putList("/project/" + projectId + "/file/translated/language/" + language.getLanguageCode() + "/document.json", TranslatedDocument.class, documents);
    }

    @Override
    public Webhook createWebhook(String eventType, String uri, Project project) throws CloudwordsClientException {
        if (project != null) {
            return createWebhook(eventType, uri, "project", project.getId(), null, null);
        } else {
            return null;
        }
    }
    
    @Override
    public Webhook createWebhook(String eventType, String uri, Project project, Language language) throws CloudwordsClientException {
        if (language != null) {
            return createWebhook(eventType, uri, "project", project.getId(), "language", language.getLanguageCode());
        } else {
            return createWebhook(eventType, uri, project);
        }
    }

    @Override
    public Webhook createWebhook(String eventType, String uri, String entityType, Integer entityId, String contextType, String contextId) throws CloudwordsClientException {
        if (eventType != null && uri != null) {
            Webhook webhook = new Webhook();
            webhook.setEventType(new CodeAndDisplay(eventType));
            webhook.setUri(uri);
            if( entityType != null && entityId != null ) {
                webhook.setEntityType(new CodeAndDisplay(entityType));
                webhook.setEntityId(entityId);
            }
            if( contextType != null && contextId != null ) {
                webhook.setContextType(new CodeAndDisplay(contextType));
                webhook.setContextId(contextId);
            }
            HttpPost request = new HttpPost(baseUrlWithVersion + "/webhook");
            setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
            try {
                request.setEntity(new StringEntity(mapper.writeValueAsString(webhook),UTF_8));
            } catch (Exception ex) {
                throw new CloudwordsClientException(ex);
            }
            return executeRequest(request, Webhook.class);
        }
        return null;
    }

    @Override
    public Webhook getWebhook(Integer webhookId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/webhook/" + webhookId + ".json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        return executeRequest(request, Webhook.class);
    }

    @Override
    public void deleteWebhook(Integer webhookId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpDelete(baseUrlWithVersion + "/webhook/" + webhookId);
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        executeRequest(request);       
    }

    @Override
    public List<Webhook> getWebhooks() throws CloudwordsClientException {
        return getList("/webhook.json", Webhook.class);
    }


    @Override
    public VendorTranslationRequest getCurrentTranslationRequestForProject(int projectId) throws CloudwordsClientException {
        HttpUriRequest request = new HttpGet(baseUrlWithVersion + "/project/" + projectId + "/translation-request/current.json");
        setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
        return executeRequest(request, VendorTranslationRequest.class);
    }

    
    @Override
    public VendorTranslationRequest acceptTranslationRequest(int projectId, VendorTranslationRequest transRequest) throws CloudwordsClientException {
        if (transRequest != null) {
            HttpPost request = new HttpPost(baseUrlWithVersion + "/project/" + projectId + "/translation-request");
            setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
            try {
                CodeAndDisplay status = new CodeAndDisplay();
                if (transRequest.isRequiresBidding()) {
                    status.setCode("vendor_accepted");
                } else {
                    status.setCode("translation_request_accepted");
                }
                transRequest.setStatus(status);
                request.setEntity(new StringEntity(mapper.writeValueAsString(transRequest), UTF_8));
            } catch (Exception ex) {
                throw new CloudwordsClientException(ex);
            }
            return executeRequest(request, VendorTranslationRequest.class);
        }
        return null;
    }

    @Override
    public VendorTranslationRequest rejectTranslationRequest(int projectId, VendorTranslationRequest transRequest) throws CloudwordsClientException {
        if (transRequest != null) {
            if (StringUtils.isEmpty(transRequest.getRejectionReason())) {
                throw new CloudwordsClientException("Rejection reason needs to be specified for rejecting the translation request.");
            }
            HttpPost request = new HttpPost(baseUrlWithVersion + "/project/" + projectId + "/translation-request");
            setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
            try {
                CodeAndDisplay status = new CodeAndDisplay();
                if (transRequest.isRequiresBidding()) {
                    status.setCode("vendor_rejected");
                } else {
                    status.setCode("translation_request_rejected");
                }
                transRequest.setStatus(status);
                request.setEntity(new StringEntity(mapper.writeValueAsString(transRequest), UTF_8));
            } catch (Exception ex) {
                throw new CloudwordsClientException(ex);
            }
            return executeRequest(request, VendorTranslationRequest.class);
        }
        return null;
    }


    @Override
    public VendorTranslationRequest submitTranslationRequest(int projectId, VendorTranslationRequest transRequest) throws CloudwordsClientException {
        if (transRequest != null) {
            if (transRequest.isRequiresBidding()) {
                if (transRequest.getBidItems() == null || transRequest.getBidItems().isEmpty()) {
                    throw new CloudwordsClientException("Translation request can't be submitted as the bidItems is either null or empty");
                }
                HttpPut request = new HttpPut(baseUrlWithVersion + "/project/" + projectId + "/translation-request");
                setBasicHeaders(request, JSON_CONTENT_TYPE, JSON_CONTENT_TYPE);
                try {
                    CodeAndDisplay status = new CodeAndDisplay();
                    status.setCode("submitted");
                    transRequest.setStatus(status);
                    request.setEntity(new StringEntity(mapper.writeValueAsString(transRequest), UTF_8));
                } catch (Exception ex) {
                    throw new CloudwordsClientException(ex);
                }
                return executeRequest(request, VendorTranslationRequest.class);
            } else {
                throw new CloudwordsClientException("Translation request object is not configured for bid submission.");
            }
        }
        return null;
    }
    
    @Override
    public PagedResponse<Project> getProjects(PagedProjectQuery query) throws CloudwordsClientException {
        return get("/project.json", new TypeReference<PagedResponse<Project>>() {}, query);
    }
}
