/*
 * Copyright 2011, Cloudwords, Inc.
 *
 * Licensed under the API LICENSE AGREEMENT, Version 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.cloudwords.com/developers/license-1.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.cloudwords.api.client.resources;

import java.util.Date;
import java.util.List;

import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;
import org.apache.commons.lang3.builder.ToStringBuilder;

/**
 * Represents a request for bids to translate the source content for a given Cloudwords
 * {@link Project} from the specified source {@link Language} to the specified target 
 * {@link Language}s. A {@link BidRequest} can specify what {@link Vendor}s to request
 * bids from, can let Cloudwords choose the best {@link Vendor}s automatically, 
 * or both.
 * 
 * @author jonathansu
 * @since 1.0
 */
public class BidRequest {

    private Date createdDate;
    private List<Vendor> preferredVendors;
    private boolean doLetCloudwordsChoose;
    private boolean doAutoSelectBidFromVendor;
    private String path;
    private Integer winningBidId;

    /**
     * Get the date in which the bid request was created. Please note that the Date object contains
     * the full time, date and time zone information.
     * <p>
     * Please refer to the API documentation for more information on date handling.
     * </p>
     * 
     * @return Date the {@link BidRequest} was created
     */
    public Date getCreatedDate() {
        return createdDate;
    }

    /**
     * Set the {@link Project} created date. (Note: {@link BidRequest} created date is a read-only
     * field with the Cloudwords API - setting it and sending it to the API will have no effect.)
     * 
     * @param createdDate
     */
    public void setCreatedDate(Date createdDate) {
        this.createdDate = createdDate;
    }

    /**
     * Get the list of preferred {@link Vendor} that are associated with this {@link BidRequest}.
     * 
     * @return The preferred {@link Vendor}s requested in this {@link BidRequest}
     */
    public List<Vendor> getPreferredVendors() {
        return preferredVendors;
    }

    /**
     * Set the list of preferred {@link Vendor} that are associated with this {@link BidRequest}.
     * 
     * @param preferredVendors
     *            The {@link Vendor}s requested for this {@link BidRequest}
     */
    public void setPreferredVendors(List<Vendor> preferredVendors) {
        this.preferredVendors = preferredVendors;
    }

    /**
     * Allow Cloudwords to automatically choose the {@link Vendor}s that best match customer's
     * {@link Project} requirements.
     * 
     * @return doLetCloudwordsChoose Whether the {@link BidRequest} automatically chooses some or
     *         all {@link Vendor}s
     */
    public boolean isDoLetCloudwordsChoose() {
        return doLetCloudwordsChoose;
    }

    /**
     * Set option to allow Cloudwords to automatically choose the {@link Vendor}s that best match
     * a customer's {@link Project} requirements.
     * 
     * @param doLetCloudwordsChoose
     *            Whether the {@link BidRequest} automatically chooses some or all {@link Vendor}s
     */
    public void setDoLetCloudwordsChoose(boolean doLetCloudwordsChoose) {
        this.doLetCloudwordsChoose = doLetCloudwordsChoose;
    }

    
    /**
     * Enables automatic selection of the submitted {@link Vendor} bid for this {@link BidRequest}.
     * This means that the {@link Project} will instantly move into translation once the
     * {@link Vendor} has submitted their bid, and therefore you will not have the ability to review
     * and reject the submitted {@link Vendor} bid. However, you can always choose to cancel the
     * {@link Project} at any time. 
     * 
     * @return doAutoSelectBidFromVendor Whether to automatically select the bid provided by the
     *         {@link Vendor} for this {@link BidRequest}
     */
    public boolean isDoAutoSelectBidFromVendor() {
        return doAutoSelectBidFromVendor;
    }

    /**
     * Set the option to automatically select the submitted {@link Vendor} bid for this
     * {@link BidRequest}. This means that the {@link Project} will instantly move into translation
     * once the {@link Vendor} has provided their bid, and therefore you will not have the ability
     * to review and reject the submitted {@link Vendor} bid. However, you can always choose to
     * cancel the {@link Project} at any time. 
     * 
     * Note: Setting This option is only possible for this {@link BidRequest} if you select a single 
     * preferred {@link Vendor} and you DO NOT have the {@link #doLetCloudwordsChoose} option enabled.
     * 
     * @param doAutoSelectBidFromVendor
     *            Whether to automatically select the bid provided by the {@link Vendor} for this
     *            {@link BidRequest}
     */
    public void setDoAutoSelectBidFromVendor(boolean doAutoSelectBidFromVendor) {
        this.doAutoSelectBidFromVendor = doAutoSelectBidFromVendor;
    }
    
    
    /**
     * Which {@link Vendor} {@link CustomerBid} was selected as the winner for this {@link BidRequest}?
     * 
     * @return winningBidId  The {@link Vendor} {@link CustomerBid} that was selected as the winner for this {@link BidRequest}
     */
    public Integer getWinningBidId() {
        return winningBidId;
    }

    /**
     * Set which {@link Vendor} {@link CustomerBid} was selected as the winner for this {@link BidRequest}?
     * 
     * @param winningBidId  
     *          The {@link Vendor} {@link CustomerBid} that was selected as the winner for this {@link BidRequest}
     */
    public void setWinningBidId(Integer winningBidId) {
        this.winningBidId = winningBidId;
    }

    /**
     * Get the API-accessible path to this {@link BidRequest}. This is the HTTP accessible path to
     * perform a GET on the {@link BidRequest}. Please refer to the API documentation for more detail.
     * 
     * @return path The URL to this {@link BidRequest}
     */
    public String getPath() {
        return path;
    }

    /**
     * Set the API-accessible path to this {@link BidRequest}. (Note: The path is a read-only field
     * with Cloudwords API - Cloudwords will automatically calculate the API-accessible path on any
     * {@link BidRequest} representation returned from the API).
     * 
     * @param path
     *            The URL to this {@link BidRequest}
     */
    public void setPath(String path) {
        this.path = path;
    }

    @Override
    public String toString() {
        return new ToStringBuilder(this).append("createdDate", createdDate)
                .append("preferredVendors", preferredVendors).append("doLetCloudwordsChoose", doLetCloudwordsChoose)
                .append("doAutoSelectBidFromVendor", doAutoSelectBidFromVendor)
                .append("winningBidId", winningBidId).append("path", path).toString();
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder().append(createdDate).append(preferredVendors).append(doLetCloudwordsChoose)
                .append(doAutoSelectBidFromVendor).append(winningBidId).append(path).hashCode();
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }
        if (obj == this) {
            return true;
        }
        if (obj.getClass() != getClass()) {
            return false;
        }
        BidRequest rhs = (BidRequest)obj;
        return new EqualsBuilder().append(createdDate, rhs.createdDate).append(preferredVendors, rhs.preferredVendors)
                .append(doLetCloudwordsChoose, rhs.doLetCloudwordsChoose)
                .append(doAutoSelectBidFromVendor, rhs.doAutoSelectBidFromVendor)
                .append(winningBidId, rhs.winningBidId).append(path, rhs.path).isEquals();
   }

}
