/*
 * Copyright 2012, Cloudwords, Inc.
 *
 * Licensed under the API LICENSE AGREEMENT, Version 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.cloudwords.com/developers/license-1.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.cloudwords.api.client.resources;

import java.math.BigDecimal;
import java.util.Date;
import java.util.List;
import java.util.Map;

import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;
import org.apache.commons.lang3.builder.ToStringBuilder;

/**
 * Represents a {@link Project} resource in Cloudwords. A {@link Project} is the central resource in
 * Cloudwords, as it represents an initiative to translate some content. It contains information,
 * both necessary and optional, to define a {@link Project}'s requirements, such as the content's
 * source {@link Language} and requested target {@link Language}s to translate into.
 * 
 * @author jonathansu
 * @since 1.0
 */
public class Project extends BaseProject {

    private String description;
    private String notes;
    private String poNumber;
    private IntendedUse intendedUse;
    private Language sourceLanguage;
    private List<Language> targetLanguages;
    private CodeAndDisplay status;
    private Date createdDate;
    private Date bidSelectDeadlineDate;
    private Date bidDueDate;
    private Date deliveryDueDate;
    private Date reviewDueDate;
    private String path;
    private BigDecimal amount;
    private Department department;
    private ProjectUpdateGlossary projectUpdateGlossary;
    private List<User> followers;
    private User owner;
    private String projectContentType;
    private Map<String,Boolean> uiFeatures;
    private String currency;

    /**
     * Get the {@link Project} description.
     * 
     * @return description of this {@link Project}
     */
    public String getDescription() {
        return description;
    }

    /**
     * Set the {@link Project} description.
     * 
     * @param description
     *            of this {@link Project}
     */
    public void setDescription(String description) {
        this.description = description;
    }

    /**
     * Get the {@link Project} notes.
     * 
     * @return notes of this {@link Project}
     */
    public String getNotes() {
        return notes;
    }

    /**
     * Set the {@link Project} notes.
     * 
     * @param notes
     *            of this {@link Project}
     */
    public void setNotes(String notes) {
        this.notes = notes;
    }

    /**
     * Get the {@link Project} {@link IntendedUse}.
     * 
     * @return {@link IntendedUse} of this {@link Project}
     */
    public IntendedUse getIntendedUse() {
        return intendedUse;
    }

    /**
     * Set the {@link Project} {@link IntendedUse}.
     * 
     * @param intendedUse
     *            of this {@link Project}
     */
    public void setIntendedUse(IntendedUse intendedUse) {
        this.intendedUse = intendedUse;
    }

    /**
     * Get the {@link Project} source {@link Language}.
     * 
     * @return sourceLanguage of this {@link Project}
     */
    public Language getSourceLanguage() {
        return sourceLanguage;
    }

    /**
     * Set the {@link Project} source {@link Language}.
     * 
     * @param sourceLanguage
     *            of this {@link Project}
     */
    public void setSourceLanguage(Language sourceLanguage) {
        this.sourceLanguage = sourceLanguage;
    }

    /**
     * Get the {@link Project} target {@link Language}(s).
     * 
     * @return targetLanguages of this {@link Project}
     */
    public List<Language> getTargetLanguages() {
        return targetLanguages;
    }

    /**
     * Set the {@link Project} target {@link Language}(s).
     * 
     * @param targetLanguages
     *            of this {@link Project}
     */
    public void setTargetLanguages(List<Language> targetLanguages) {
        this.targetLanguages = targetLanguages;
    }

    /**
     * Get the {@link Project} status.
     * 
     * @return status of this {@link Project}
     */
    public CodeAndDisplay getStatus() {
        return status;
    }

    /**
     * Set the {@link Project} status. (Note: Project status is a read-only
     * field with the Cloudwords API - setting a {@link Project} status and sending it to the API
     * will have no effect.)
     * 
     * @param status
     *            of this {@link Project}
     */
    public void setStatus(CodeAndDisplay status) {
        this.status = status;
    }

    /**
     * Get the {@link Project} purchase order number.
     * 
     * @return poNumber of this {@link Project}
     */
    public String getPoNumber() {
        return poNumber;
    }

    /**
     * Set the {@link Project} purchase order number.
     * 
     * @param poNumber
     *            of this {@link Project}
     */
    public void setPoNumber(String poNumber) {
        this.poNumber = poNumber;
    }

    /**
     * Get the {@link Project} created date. Please note that the {@link Date} object contains the
     * full time, date and time zone information.
     * <p>
     * Please refer to the API documentation for more information on date handling.
     * </p>
     * 
     * @return createdDate of this {@link Project}
     */
    public Date getCreatedDate() {
        return createdDate;
    }

    /**
     * Set the {@link Project} created date. (Note: Project created date is a read-only field with
     * the Cloudwords API - setting it and sending it to the API will have no effect.)
     * 
     * @param createdDate
     *            of this {@link Project}
     */
    public void setCreatedDate(Date createdDate) {
        this.createdDate = createdDate;
    }

    /**
     * Get the {@link Project} bid due date. Please note that the {@link Date} object contains the
     * full time, date and time zone information.
     * <p>
     * Please refer to the API documentation for more information on date handling.
     * </p>
     * 
     * @return bidDueDate of this {@link Project}
     */
    public Date getBidDueDate() {
        return bidDueDate;
    }

    /**
     * Set the {@link Project} bid due date. Please note that the {@link Date} object contains the
     * full time, date and time zone information that Cloudwords API will respect.
     * <p>
     * Please refer to the API documentation for more information on date handling.
     * </p>
     * 
     * @param bidDueDate
     */
    public void setBidDueDate(Date bidDueDate) {
        this.bidDueDate = bidDueDate;
    }

    /**
     * Get the {@link Project} bid selection deadline. Please note that the {@link Date} object
     * contains the full time, date and time zone information.
     * <p>
     * Please refer to the API documentation for more information on date handling.
     * </p>
     * 
     * @return bidSelectDeadlineDate
     */
    public Date getBidSelectDeadlineDate() {
        return bidSelectDeadlineDate;
    }

    /**
     * Set the {@link Project} bid selection deadline. (Note: Project bid selection deadline is a
     * read-only field with the Cloudwords API - Cloudwords will automatically calculate it based on
     * the specified bid due date and {@link Project} due date.)
     * <p>
     * Please refer to the API documentation for more information on date handling.
     * </p>
     * 
     * @param bidSelectDeadlineDate
     */
    public void setBidSelectDeadlineDate(Date bidSelectDeadlineDate) {
        this.bidSelectDeadlineDate = bidSelectDeadlineDate;
    }

    /**
     * Get the {@link Project} delivery due date. Please note that the {@link Date} object contains the
     * full time, date and time zone information.
     * <p>
     * Please refer to the API documentation for more information on date handling.
     * </p>
     * 
     * @return deliveryDueDate
     */
    public Date getDeliveryDueDate() {
        return deliveryDueDate;
    }

    /**
     * Set the {@link Project} delivery due date. Please note that the {@link Date} object contains the
     * full time, date and time zone information that Cloudwords API will respect.
     * <p>
     * Please refer to the API documentation for more information on date handling.
     * </p>
     * 
     * @param deliveryDueDate
     */
    public void setDeliveryDueDate(Date deliveryDueDate) {
        this.deliveryDueDate = deliveryDueDate;
    }

    
    /**
     * Get the amount or cost associated with this {@link Project}. 
     * 
     * @return amount or cost of this {@link Project}. 
     */
    public BigDecimal getAmount() {
        return amount;
    }

    
    /**
     * Set the amount or cost associated with this {@link Project}. (Note: Project amount is a read-only
     * field with the Cloudwords API - setting an amount and sending it to the API
     * will have no effect.)
     * 
     * @param amount
     *            amount or cost of this {@link Project}.
     */
    public void setAmount(BigDecimal amount) {
        this.amount = amount;
    }
    
    /**
     * Get the {@link Customer} {@link Department} that this {@link Project} belongs to.
     * 
     * @return the {@link Department} information for this project.
     */
    public Department getDepartment() {
        return department;
    }

    /**
     * set the {@link Customer} {@link Department} that this {@link Project} belongs to.
     * @param department
     */
    public void setDepartment(Department department) {
        this.department = department;
    }
    

    @Override
    public String toString() {
        return new ToStringBuilder(this).append("id", getId()).append("name", getName()).append("description", description)
                .append("notes", notes).append("poNumber", poNumber).append("intendedUse", intendedUse)
                .append("sourceLanguage", sourceLanguage).append("targetLanguages", targetLanguages)
                .append("status", status).append("createdDate", createdDate)
                .append("bidSelectDeadlineDate", bidSelectDeadlineDate).append("bidDueDate", bidDueDate)
                .append("deliveryDueDate", deliveryDueDate).append("amount", amount).append("path", path).append("reviewDueDate", reviewDueDate).append("projectContentType", projectContentType).toString();
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder().append(getId()).append(getName()).append(description).append(notes).append(poNumber)
                .append(intendedUse).append(sourceLanguage).append(targetLanguages).append(status).append(createdDate)
                .append(bidSelectDeadlineDate).append(bidDueDate).append(deliveryDueDate).append(amount).append(path).append(reviewDueDate).append(projectContentType).hashCode();
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }
        if (obj == this) {
            return true;
        }
        if (obj.getClass() != getClass()) {
            return false;
        }
        Project rhs = (Project)obj;
        return new EqualsBuilder().append(getId(), rhs.getId()).append(getName(), rhs.getName()).append(description, rhs.description)
                .append(notes, rhs.notes).append(poNumber, rhs.poNumber).append(intendedUse, rhs.intendedUse)
                .append(sourceLanguage, rhs.sourceLanguage).append(targetLanguages, rhs.targetLanguages)
                .append(status, rhs.status).append(createdDate, rhs.createdDate)
                .append(bidSelectDeadlineDate, rhs.bidSelectDeadlineDate).append(bidDueDate, rhs.bidDueDate)
                .append(deliveryDueDate, rhs.deliveryDueDate).append(amount, rhs.amount).append(path, rhs.path).append(reviewDueDate, rhs.reviewDueDate).append(projectContentType, rhs.projectContentType).isEquals();
    }

    public ProjectUpdateGlossary getProjectUpdateGlossary() {
        return projectUpdateGlossary;
    }

    public void setProjectUpdateGlossary(ProjectUpdateGlossary projectUpdateGlossary) {
        this.projectUpdateGlossary = projectUpdateGlossary;
    }

    public User getOwner() {
        return owner;
    }

    public void setOwner(User owner) {
        this.owner = owner;
    }

    public List<User> getFollowers() {
        return followers;
    }

    public void setFollowers(List<User> followers) {
        this.followers = followers;
    }

    public Date getReviewDueDate() {
        return reviewDueDate;
    }

    public void setReviewDueDate(Date reviewDueDate) {
        this.reviewDueDate = reviewDueDate;
    }

    /**
     * The content type of the Project. For projects retrieved through the API, this may be a reserved type such
     * as STANDARD, SFDC_KB, MARKETO, CW_GLOSSARY or MARKETO. For projects created through the API, this will be
     * STANDARD or a custom name specified by the API integration.
     */
    public String getProjectContentType() {
        return projectContentType;
    }

    /**
     * This value can only be set when the project is created, and cannot be changed thereafter. Valid values
     * when creating a project through the API are STANDARD, or any string that is not reserved by Cloudwords (see {@link #getProjectContentType()}).
     * Custom values specified will be displayed to the user in the Cloudwords UI and can also be used to filter projects
     * on the project search UI. It is suggested that integrators use a string that is specific to that integration (such
     * as the integration name).
     */
    public void setProjectContentType(String projectContentType) {
        this.projectContentType = projectContentType;
    }

    /**
     * Get the set of UI Features, and the 'permitted' value for each (that is, whether the Cloudwords UI will allow
     * that action). See the Cloudwords Developer Center for a list of valid UI Features.
     */
    public Map<String,Boolean> getUiFeatures() {
        return uiFeatures;
    }

    public void setUiFeatures(Map<String,Boolean> uiFeatures) {
        this.uiFeatures = uiFeatures;
    }

    /**
     * Get the currency code for this project.
     */
    public String getCurrency() {
        return currency;
    }

    public void setCurrency(String currency) {
        this.currency = currency;
    }

}
